# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#               2015 - Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import itertools
import math
import statistics
import sys
from . import mi
from . import termgraph
from .command import Command
from ..core import irq as core_irq
from ..linuxautomaton import sv


class IrqAnalysisCommand(Command):
    _DESC = """The irq command."""
    _ANALYSIS_CLASS = core_irq.IrqAnalysis
    _MI_TITLE = 'System interrupt analysis'
    _MI_DESCRIPTION = 'Interrupt frequency distribution, statistics, and log'
    _MI_TAGS = [mi.Tags.INTERRUPT, mi.Tags.STATS, mi.Tags.FREQ, mi.Tags.LOG]
    _MI_TABLE_CLASS_LOG = 'log'
    _MI_TABLE_CLASS_HARD_STATS = 'hard-stats'
    _MI_TABLE_CLASS_SOFT_STATS = 'soft-stats'
    _MI_TABLE_CLASS_FREQ = 'freq'
    _MI_TABLE_CLASS_SUMMARY = 'summary'
    _MI_TABLE_CLASSES = [
        (
            _MI_TABLE_CLASS_LOG,
            'Interrupt log', [
                ('time_range', 'Time range', mi.TimeRange),
                ('raised_ts', 'Raised timestamp', mi.Timestamp),
                ('cpu', 'CPU', mi.Cpu),
                ('irq', 'Interrupt', mi.Irq),
            ]
        ),
        (
            _MI_TABLE_CLASS_HARD_STATS,
            'Hardware interrupt statistics', [
                ('irq', 'Interrupt', mi.Irq),
                ('count', 'Interrupt count', mi.Number, 'interrupts'),
                ('min_duration', 'Minimum duration', mi.Duration),
                ('avg_duration', 'Average duration', mi.Duration),
                ('max_duration', 'Maximum duration', mi.Duration),
                ('stdev_duration', 'Interrupt duration standard deviation',
                 mi.Duration),
            ]
        ),
        (
            _MI_TABLE_CLASS_SOFT_STATS,
            'Software interrupt statistics', [
                ('irq', 'Interrupt', mi.Irq),
                ('count', 'Interrupt count', mi.Number, 'interrupts'),
                ('min_duration', 'Minimum duration', mi.Duration),
                ('avg_duration', 'Average duration', mi.Duration),
                ('max_duration', 'Maximum duration', mi.Duration),
                ('stdev_duration', 'Interrupt duration standard deviation',
                 mi.Duration),
                ('raise_count', 'Interrupt raise count', mi.Number,
                 'interrupt raises'),
                ('min_latency', 'Minimum raise latency', mi.Duration),
                ('avg_latency', 'Average raise latency', mi.Duration),
                ('max_latency', 'Maximum raise latency', mi.Duration),
                ('stdev_latency', 'Interrupt raise latency standard deviation',
                 mi.Duration),
            ]
        ),
        (
            _MI_TABLE_CLASS_FREQ,
            'Interrupt handler duration frequency distribution', [
                ('duration_lower', 'Duration (lower bound)', mi.Duration),
                ('duration_upper', 'Duration (upper bound)', mi.Duration),
                ('count', 'Interrupt count', mi.Number, 'interrupts'),
            ]
        ),
        (
            _MI_TABLE_CLASS_SUMMARY,
            'Interrupt statistics - summary', [
                ('time_range', 'Time range', mi.TimeRange),
                ('count', 'Total interrupt count', mi.Number, 'interrupts'),
            ]
        ),
    ]

    def _analysis_tick(self, period_data, end_ns):
        if period_data is None:
            return

        begin_ns = period_data.period.begin_evt.timestamp
        log_table = None
        hard_stats_table = None
        soft_stats_table = None
        freq_tables = None

        if self._args.log:
            log_table = self._get_log_result_table(period_data,
                                                   begin_ns, end_ns)

        if self._args.stats or self._args.freq:
            hard_stats_table, soft_stats_table, freq_tables = \
                self._get_stats_freq_result_tables(period_data,
                                                   begin_ns, end_ns)

        if self._mi_mode:
            self._mi_append_result_table(log_table)
            self._mi_append_result_table(hard_stats_table)
            self._mi_append_result_table(soft_stats_table)

            if self._args.freq_series:
                freq_tables = [self._get_freq_series_table(freq_tables)]

            self._mi_append_result_tables(freq_tables)
        else:
            self._print_date(begin_ns, end_ns)

            if hard_stats_table or soft_stats_table or freq_tables:
                self._print_stats_freq(hard_stats_table, soft_stats_table,
                                       freq_tables)
                if log_table:
                    print()

            if log_table:
                self._print_log(log_table)

    def _create_summary_result_tables(self):
        if not self._args.stats:
            self._mi_clear_result_tables()
            return

        hard_stats_tables = \
            self._mi_get_result_tables(self._MI_TABLE_CLASS_HARD_STATS)
        soft_stats_tables = \
            self._mi_get_result_tables(self._MI_TABLE_CLASS_SOFT_STATS)
        assert len(hard_stats_tables) == len(soft_stats_tables)
        begin = hard_stats_tables[0].timerange.begin.value
        end = hard_stats_tables[-1].timerange.end.value
        summary_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_SUMMARY,
                                         begin, end)

        for hs_table, ss_table in zip(hard_stats_tables, soft_stats_tables):
            assert hs_table.timerange == ss_table.timerange

            for row in itertools.chain(hs_table.rows, ss_table.rows):
                summary_table.append_row(
                    time_range=hs_table.timerange,
                    count=row.count,
                )

        self._mi_clear_result_tables()
        self._mi_append_result_table(summary_table)

    def _get_log_result_table(self, period_data, begin_ns, end_ns):
        result_table = self._mi_create_result_table(self._MI_TABLE_CLASS_LOG,
                                                    begin_ns, end_ns)

        for irq in period_data.irq_list:
            if not self._filter_irq(irq):
                continue

            if type(irq) is sv.HardIRQ:
                is_hard = True
                raised_ts_do = mi.Empty()
                name = period_data.hard_irq_stats[irq.id].name
            else:
                is_hard = False

                if irq.raise_ts is None:
                    raised_ts_do = mi.Unknown()
                else:
                    raised_ts_do = mi.Timestamp(irq.raise_ts)

                name = period_data.softirq_stats[irq.id].name

            result_table.append_row(
                time_range=mi.TimeRange(irq.begin_ts, irq.end_ts),
                raised_ts=raised_ts_do,
                cpu=mi.Cpu(irq.cpu_id),
                irq=mi.Irq(is_hard, irq.id, name),
            )

        return result_table

    def _get_common_stats_result_table_row(self, is_hard, irq_nr, irq_stats):
        stdev = self._compute_duration_stdev(irq_stats)

        if math.isnan(stdev):
            stdev = mi.Unknown()
        else:
            stdev = mi.Duration(stdev)

        return (
            mi.Irq(is_hard, irq_nr, irq_stats.name),
            mi.Number(irq_stats.count),
            mi.Duration(irq_stats.min_duration),
            mi.Duration(irq_stats.total_duration / irq_stats.count),
            mi.Duration(irq_stats.max_duration),
            stdev,
        )

    def _append_hard_stats_result_table_row(self, irq_nr, irq_stats,
                                            hard_stats_table):
        common_row = self._get_common_stats_result_table_row(True, irq_nr,
                                                             irq_stats)
        hard_stats_table.append_row(
            irq=common_row[0],
            count=common_row[1],
            min_duration=common_row[2],
            avg_duration=common_row[3],
            max_duration=common_row[4],
            stdev_duration=common_row[5],
        )

    def _append_soft_stats_result_table_row(self, irq_nr, irq_stats,
                                            soft_stats_table):
        common_row = self._get_common_stats_result_table_row(False, irq_nr,
                                                             irq_stats)

        if irq_stats.raise_count == 0:
            min_latency = mi.Unknown()
            avg_latency = mi.Unknown()
            max_latency = mi.Unknown()
            stdev_latency = mi.Unknown()
        else:
            min_latency = mi.Duration(irq_stats.min_raise_latency)
            avg_latency = irq_stats.total_raise_latency / irq_stats.raise_count
            avg_latency = mi.Duration(avg_latency)
            max_latency = mi.Duration(irq_stats.max_raise_latency)
            stdev = self._compute_raise_latency_stdev(irq_stats)

            if math.isnan(stdev):
                stdev_latency = mi.Unknown()
            else:
                stdev_latency = mi.Duration(stdev)

        soft_stats_table.append_row(
            irq=common_row[0],
            count=common_row[1],
            min_duration=common_row[2],
            avg_duration=common_row[3],
            max_duration=common_row[4],
            stdev_duration=common_row[5],
            raise_count=mi.Number(irq_stats.raise_count),
            min_latency=min_latency,
            avg_latency=avg_latency,
            max_latency=max_latency,
            stdev_latency=stdev_latency,
        )

    def _fill_freq_result_table(self, period_data, irq_stats, freq_table):
        # The number of bins for the histogram
        resolution = self._args.freq_resolution
        if self._args.min is not None:
            min_duration = self._args.min
        else:
            min_duration = irq_stats.min_duration
        if self._args.max is not None:
            max_duration = self._args.max
        else:
            max_duration = irq_stats.max_duration

        # ns to µs
        min_duration /= 1000
        max_duration /= 1000

        # histogram's step
        if self._args.freq_uniform:
            # TODO: perform only one time
            durations = [irq.duration for irq in period_data.irq_list]
            min_duration, max_duration, step = \
                self._find_uniform_freq_values(durations)
        else:
            step = (max_duration - min_duration) / resolution

        if step == 0:
            return

        buckets = []
        counts = []

        for i in range(resolution):
            buckets.append(i * step)
            counts.append(0)

        for irq in irq_stats.irq_list:
            duration = irq.duration / 1000
            index = int((duration - min_duration) / step)

            if index >= resolution:
                # special case for max value: put in last bucket (includes
                # its upper bound)
                if duration == max_duration:
                    counts[index - 1] += 1

                continue

            counts[index] += 1

        for index, count in enumerate(counts):
            lower_bound = index * step + min_duration
            upper_bound = (index + 1) * step + min_duration
            freq_table.append_row(
                duration_lower=mi.Duration.from_us(lower_bound),
                duration_upper=mi.Duration.from_us(upper_bound),
                count=mi.Number(count),
            )

    def _fill_stats_freq_result_tables(self, period_data, begin_ns,
                                       end_ns, is_hard,
                                       analysis_stats, filter_list,
                                       hard_stats_table, soft_stats_table,
                                       freq_tables):
        for id in sorted(analysis_stats):
            if filter_list and str(id) not in filter_list:
                continue

            irq_stats = analysis_stats[id]

            if irq_stats.count == 0:
                continue

            if self._args.stats:
                if is_hard:
                    append_row_fn = self._append_hard_stats_result_table_row
                    table = hard_stats_table
                else:
                    append_row_fn = self._append_soft_stats_result_table_row
                    table = soft_stats_table

                append_row_fn(id, irq_stats, table)

            if self._args.freq:
                subtitle = '{} ({})'.format(irq_stats.name, id)
                freq_table = \
                    self._mi_create_result_table(self._MI_TABLE_CLASS_FREQ,
                                                 begin_ns, end_ns, subtitle)
                self._fill_freq_result_table(period_data, irq_stats,
                                             freq_table)

                # it is possible that the frequency distribution result
                # table is empty; we need to keep it any way because
                # there's a 1-to-1 association between the statistics
                # row indexes (if available) and the frequency table
                # indexes
                freq_tables.append(freq_table)

    def _get_freq_series_table(self, freq_tables):
        if not freq_tables:
            return

        column_infos = [
            ('duration_lower', 'Duration (lower bound)', mi.Duration),
            ('duration_upper', 'Duration (upper bound)', mi.Duration),
        ]

        for index, freq_table in enumerate(freq_tables):
            column_infos.append((
                'irq{}'.format(index),
                freq_table.subtitle,
                mi.Number,
                'interrupts'
            ))

        title = 'Interrupt handlers duration frequency distributions'
        table_class = mi.TableClass(None, title, column_infos)
        begin = freq_tables[0].timerange.begin.value
        end = freq_tables[0].timerange.end.value
        result_table = mi.ResultTable(table_class, begin, end)

        for row_index, freq0_row in enumerate(freq_tables[0].rows):
            row_tuple = [
                freq0_row.duration_lower,
                freq0_row.duration_upper,
            ]

            for freq_table in freq_tables:
                freq_row = freq_table.rows[row_index]
                row_tuple.append(freq_row.count)

            result_table.append_row_tuple(tuple(row_tuple))

        return result_table

    def _get_stats_freq_result_tables(self, period_data, begin_ns, end_ns):
        def fill_stats_freq_result_tables(period_data, is_hard,
                                          stats, filter_list):
            self._fill_stats_freq_result_tables(period_data, begin_ns, end_ns,
                                                is_hard, stats, filter_list,
                                                hard_stats_table,
                                                soft_stats_table, freq_tables)

        hard_stats_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_HARD_STATS,
                                         begin_ns, end_ns)
        soft_stats_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_SOFT_STATS,
                                         begin_ns, end_ns)
        freq_tables = []

        if self._args.irq_filter_list is not None or \
           self._args.softirq_filter_list is None:
            fill_stats_freq_result_tables(period_data, True,
                                          period_data.hard_irq_stats,
                                          self._args.irq_filter_list)

        if self._args.softirq_filter_list is not None or \
           self._args.irq_filter_list is None:
            fill_stats_freq_result_tables(period_data, False,
                                          period_data.softirq_stats,
                                          self._args.softirq_filter_list)

        return hard_stats_table, soft_stats_table, freq_tables

    def _print_log(self, result_table):
        fmt = '[{:<18}, {:<18}] {:>15} {:>4}  {:<9} {:>4}  {:<22}'
        title_fmt = '{:<20} {:<19} {:>15} {:>4}  {:<9} {:>4}  {:<22}'
        print(title_fmt.format('Begin', 'End', 'Duration (us)', 'CPU',
                               'Type', '#', 'Name'))
        for row in result_table.rows:
            timerange = row.time_range
            begin_ts = timerange.begin.value
            end_ts = timerange.end.value

            if type(row.raised_ts) is mi.Timestamp:
                raised_ts = ' (raised at {})'.format(
                    self._format_timestamp(row.raised_ts.value)
                )
            else:
                raised_ts = ''

            cpu_id = row.cpu.id
            irq_do = row.irq

            if irq_do.is_hard:
                irqtype = 'IRQ'
            else:
                irqtype = 'SoftIRQ'

            print(fmt.format(self._format_timestamp(begin_ts),
                             self._format_timestamp(end_ts),
                             '%0.03f' % ((end_ts - begin_ts) / 1000),
                             '%d' % cpu_id, irqtype, irq_do.nr,
                             irq_do.name + raised_ts))

    def _validate_transform_args(self):
        args = self._args
        args.irq_filter_list = None
        args.softirq_filter_list = None

        if args.irq:
            args.irq_filter_list = args.irq.split(',')
        if args.softirq:
            args.softirq_filter_list = args.softirq.split(',')

    def _compute_duration_stdev(self, irq_stats_item):
        if irq_stats_item.count < 2:
            return float('nan')

        durations = []
        for irq in irq_stats_item.irq_list:
            durations.append(irq.end_ts - irq.begin_ts)

        return statistics.stdev(durations)

    def _compute_raise_latency_stdev(self, irq_stats_item):
        if irq_stats_item.raise_count < 2:
            return float('nan')

        raise_latencies = []
        for irq in irq_stats_item.irq_list:
            if irq.raise_ts is None:
                continue

            raise_latencies.append(irq.begin_ts - irq.raise_ts)

        return statistics.stdev(raise_latencies)

    def _print_frequency_distribution(self, freq_table):
        title_fmt = 'Handler duration frequency distribution {}'

        graph = termgraph.FreqGraph(
            data=freq_table.rows,
            get_value=lambda row: row.count.value,
            get_lower_bound=lambda row: row.duration_lower.to_us(),
            title=title_fmt.format(freq_table.subtitle),
            unit='µs'
        )

        graph.print_graph()

    def _filter_irq(self, irq):
        if type(irq) is sv.HardIRQ:
            if self._args.irq_filter_list:
                return str(irq.id) in self._args.irq_filter_list
            if self._args.softirq_filter_list:
                return False
        else:  # SoftIRQ
            if self._args.softirq_filter_list:
                return str(irq.id) in self._args.softirq_filter_list
            if self._args.irq_filter_list:
                return False

        return True

    def _print_hard_irq_stats_row(self, row):
        output_str = self._get_duration_stats_str(row)
        print(output_str)

    def _print_soft_irq_stats_row(self, row):
        output_str = self._get_duration_stats_str(row)

        if row.raise_count.value != 0:
            output_str += self._get_raise_latency_str(row)

        print(output_str)

    def _get_duration_stats_str(self, row):
        format_str = '{:<3} {:<18} {:>5} {:>12} {:>12} {:>12} {:>12} {:<2}'
        irq_do = row.irq
        count = row.count.value
        min_duration = row.min_duration.to_us()
        avg_duration = row.avg_duration.to_us()
        max_duration = row.max_duration.to_us()

        if type(row.stdev_duration) is mi.Unknown:
            duration_stdev_str = '?'
        else:
            duration_stdev_str = '%0.03f' % row.stdev_duration.to_us()

        output_str = format_str.format('%d:' % irq_do.nr,
                                       '<%s>' % irq_do.name,
                                       '%d' % count,
                                       '%0.03f' % min_duration,
                                       '%0.03f' % avg_duration,
                                       '%0.03f' % max_duration,
                                       '%s' % duration_stdev_str,
                                       ' |')
        return output_str

    def _get_raise_latency_str(self, row):
        format_str = ' {:>6} {:>12} {:>12} {:>12} {:>12}'
        raise_count = row.raise_count.value
        min_raise_latency = row.min_latency.to_us()
        avg_raise_latency = row.avg_latency.to_us()
        max_raise_latency = row.max_latency.to_us()

        if type(row.stdev_latency) is mi.Unknown:
            raise_latency_stdev_str = '?'
        else:
            raise_latency_stdev_str = '%0.03f' % row.stdev_latency.to_us()

        output_str = format_str.format(raise_count,
                                       '%0.03f' % min_raise_latency,
                                       '%0.03f' % avg_raise_latency,
                                       '%0.03f' % max_raise_latency,
                                       '%s' % raise_latency_stdev_str)

        return output_str

    def _print_stats_freq(self, hard_stats_table, soft_stats_table,
                          freq_tables):
        hard_header_format = '{:<52} {:<12}\n' \
                             '{:<22} {:<14} {:<12} {:<12} {:<10} {:<12}\n'
        hard_header = hard_header_format.format(
            'Hard IRQ', 'Duration (us)',
            '', 'count', 'min', 'avg', 'max', 'stdev'
        )
        hard_header += ('-' * 82 + '|')
        soft_header_format = '{:<52} {:<52} {:<12}\n' \
                             '{:<22} {:<14} {:<12} {:<12} {:<10} {:<4} ' \
                             '{:<3} {:<14} {:<12} {:<12} {:<10} {:<12}\n'
        soft_header = soft_header_format.format(
            'Soft IRQ', 'Duration (us)',
            'Raise latency (us)', '',
            'count', 'min', 'avg', 'max', 'stdev', ' |',
            'count', 'min', 'avg', 'max', 'stdev'
        )
        soft_header += '-' * 82 + '|' + '-' * 60

        if hard_stats_table.rows or soft_stats_table.rows:
            stats_rows = itertools.chain(hard_stats_table.rows,
                                         soft_stats_table.rows)

            if freq_tables:
                for stats_row, freq_table in zip(stats_rows, freq_tables):
                    irq = stats_row.irq

                    if irq.is_hard:
                        print(hard_header)
                        self._print_hard_irq_stats_row(stats_row)
                    else:
                        print(soft_header)
                        self._print_soft_irq_stats_row(stats_row)

                    # frequency table might be empty: do not print
                    if freq_table.rows:
                        print()
                        self._print_frequency_distribution(freq_table)

                    print()

            else:
                hard_header_printed = False
                soft_header_printed = False

                for stats_row in stats_rows:
                    irq = stats_row.irq

                    if irq.is_hard:
                        if not hard_header_printed:
                            print(hard_header)
                            hard_header_printed = True

                        self._print_hard_irq_stats_row(stats_row)
                    else:
                        if not soft_header_printed:
                            if hard_header_printed:
                                print()

                            print(soft_header)
                            soft_header_printed = True

                        self._print_soft_irq_stats_row(stats_row)

            return

        for freq_table in freq_tables:
            # frequency table might be empty: do not print
            if freq_table.rows:
                print()
                self._print_frequency_distribution(freq_table)

    def _add_arguments(self, ap):
        Command._add_min_max_args(ap)
        Command._add_freq_args(
            ap, help='Output the frequency distribution of handler durations')
        Command._add_log_args(
            ap, help='Output the IRQs in chronological order')
        Command._add_stats_args(ap, help='Output IRQ statistics')
        ap.add_argument('--irq', type=str, default=None,
                        help='Output results only for the list of IRQ')
        ap.add_argument('--softirq', type=str, default=None,
                        help='Output results only for the list of SoftIRQ')


def _run(mi_mode):
    irqcmd = IrqAnalysisCommand(mi_mode=mi_mode)
    irqcmd.run()


def _runstats(mi_mode):
    sys.argv.insert(1, '--stats')
    _run(mi_mode)


def _runlog(mi_mode):
    sys.argv.insert(1, '--log')
    _run(mi_mode)


def _runfreq(mi_mode):
    sys.argv.insert(1, '--freq')
    _run(mi_mode)


def runstats():
    _runstats(mi_mode=False)


def runlog():
    _runlog(mi_mode=False)


def runfreq():
    _runfreq(mi_mode=False)


def runstats_mi():
    _runstats(mi_mode=True)


def runlog_mi():
    _runlog(mi_mode=True)


def runfreq_mi():
    _runfreq(mi_mode=True)



# The MIT License (MIT)
#
# Copyright (C) 2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import stats
from .analysis import Analysis, PeriodData
from ..linuxautomaton import sv


class _PeriodData(PeriodData):
    def __init__(self):
        self.disks = {}
        self.ifaces = {}
        self.tids = {}


class IoAnalysis(Analysis):
    def __init__(self, state, conf):
        notification_cbs = {
            'net_dev_xmit': self._process_net_dev_xmit,
            'netif_receive_skb': self._process_netif_receive_skb,
            'block_rq_complete': self._process_block_rq_complete,
            'io_rq_exit': self._process_io_rq_exit,
            'create_fd': self._process_create_fd,
            'close_fd': self._process_close_fd,
            'update_fd': self._process_update_fd,
            'create_parent_proc': self._process_create_parent_proc,
            'lttng_statedump_block_device': self._process_statedump_block
        }

        super().__init__(state, conf, notification_cbs)
        if conf.cpu_list is not None:
            print('Warning: cpu filter not enabled on I/O analysis')

    def process_event(self, ev):
        super().process_event(ev)
        self._process_event_cb(ev)

    def _create_period_data(self):
        return _PeriodData()

    @property
    def disk_io_requests(self, period_data):
        for disk in period_data.disks.values():
            for io_rq in disk.rq_list:
                yield io_rq

    def io_requests(self, period_data):
        return self._get_io_requests(period_data)

    def open_io_requests(self, period_data):
        return self._get_io_requests(period_data, sv.IORequest.OP_OPEN)

    def read_io_requests(self, period_data):
        return self._get_io_requests(period_data, sv.IORequest.OP_READ)

    def write_io_requests(self, period_data):
        return self._get_io_requests(period_data, sv.IORequest.OP_WRITE)

    def close_io_requests(self, period_data):
        return self._get_io_requests(period_data, sv.IORequest.OP_CLOSE)

    def sync_io_requests(self, period_data):
        return self._get_io_requests(period_data, sv.IORequest.OP_SYNC)

    def read_write_io_requests(self, period_data):
        return self._get_io_requests(period_data, sv.IORequest.OP_READ_WRITE)

    def _get_io_requests(self, period_data, io_operation=None):
        """Create a generator of syscall io requests by operation.

        Args:
            io_operation (IORequest.OP_*, optional): The operation of
            the io_requests to return. Return all IO requests if None.
        """
        for proc in period_data.tids.values():
            for io_rq in proc.rq_list:
                if isinstance(io_rq, sv.BlockIORequest):
                    continue

                if io_operation is None or \
                   sv.IORequest.is_equivalent_operation(io_operation,
                                                        io_rq.operation):
                    yield io_rq

    def get_files_stats(self, period_data):
        files_stats = {}

        for proc_stats in period_data.tids.values():
            for fd_list in proc_stats.fds.values():
                for fd_stats in fd_list:
                    filename = fd_stats.filename
                    # Add process name to generic filenames to
                    # distinguish them
                    if FileStats.is_generic_name(filename):
                        filename += ' (%s)' % proc_stats.comm

                    if filename not in files_stats:
                        files_stats[filename] = FileStats(filename)

                    files_stats[filename].update_stats(fd_stats, proc_stats)

        return files_stats

    @staticmethod
    def _assign_fds_to_parent(proc, parent):
        if proc.fds:
            toremove = []
            for fd in proc.fds:
                if fd not in parent.fds:
                    parent.fds[fd] = proc.fds[fd]
                else:
                    # best effort to fix the filename
                    if not parent.get_fd(fd).filename:
                        parent.get_fd(fd).filename = proc.get_fd(fd).filename
                toremove.append(fd)
            for fd in toremove:
                del proc.fds[fd]

    def _process_net_dev_xmit(self, period_data, **kwargs):
        name = kwargs['iface_name']
        sent_bytes = kwargs['sent_bytes']

        if name not in period_data.ifaces:
            period_data.ifaces[name] = IfaceStats(name)

        period_data.ifaces[name].sent_packets += 1
        period_data.ifaces[name].sent_bytes += sent_bytes

    def _process_netif_receive_skb(self, period_data, **kwargs):
        name = kwargs['iface_name']
        recv_bytes = kwargs['recv_bytes']

        if name not in period_data.ifaces:
            period_data.ifaces[name] = IfaceStats(name)

        period_data.ifaces[name].recv_packets += 1
        period_data.ifaces[name].recv_bytes += recv_bytes

    def _process_block_rq_complete(self, period_data, **kwargs):
        req = kwargs['req']
        proc = kwargs['proc']
        disk = kwargs['disk']

        if disk.dev not in period_data.disks:
            period_data.disks[disk.dev] = DiskStats.new_from_disk(disk)

        period_data.disks[disk.dev].update_stats(req)

        if proc is not None:
            if proc.tid not in period_data.tids:
                period_data.tids[proc.tid] = ProcessIOStats.new_from_process(
                    proc)

            period_data.tids[proc.tid].update_block_stats(req)

    def _process_io_rq_exit(self, period_data, **kwargs):
        proc = kwargs['proc']
        parent_proc = kwargs['parent_proc']
        io_rq = kwargs['io_rq']

        if proc.tid not in period_data.tids:
            period_data.tids[proc.tid] = ProcessIOStats.new_from_process(proc)

        if parent_proc.tid not in period_data.tids:
            period_data.tids[parent_proc.tid] = (
                ProcessIOStats.new_from_process(parent_proc))

        proc_stats = period_data.tids[proc.tid]
        parent_stats = period_data.tids[parent_proc.tid]

        fd_types = {}
        if io_rq.errno is None:
            if io_rq.operation == sv.IORequest.OP_READ or \
               io_rq.operation == sv.IORequest.OP_WRITE:
                if parent_stats.get_fd(io_rq.fd) is None:
                    return
                fd_types['fd'] = parent_stats.get_fd(io_rq.fd).fd_type
            elif io_rq.operation == sv.IORequest.OP_READ_WRITE:
                if parent_stats.get_fd(io_rq.fd_in) is None:
                    return
                if parent_stats.get_fd(io_rq.fd_out) is None:
                    return
                fd_types['fd_in'] = parent_stats.get_fd(io_rq.fd_in).fd_type
                fd_types['fd_out'] = parent_stats.get_fd(io_rq.fd_out).fd_type

        proc_stats.update_io_stats(io_rq, fd_types)
        parent_stats.update_fd_stats(io_rq)

        # Check if the proc stats comm corresponds to the actual
        # process comm. It might be that it was missing so far.
        if proc_stats.comm != proc.comm:
            proc_stats.comm = proc.comm
        if parent_stats.comm != parent_proc.comm:
            parent_stats.comm = parent_proc.comm

    def _process_create_parent_proc(self, period_data, **kwargs):
        proc = kwargs['proc']
        parent_proc = kwargs['parent_proc']

        if proc.tid not in period_data.tids:
            period_data.tids[proc.tid] = ProcessIOStats.new_from_process(proc)

        if parent_proc.tid not in period_data.tids:
            period_data.tids[parent_proc.tid] = (
                ProcessIOStats.new_from_process(parent_proc))

        proc_stats = period_data.tids[proc.tid]
        parent_stats = period_data.tids[parent_proc.tid]

        proc_stats.pid = parent_stats.tid
        IoAnalysis._assign_fds_to_parent(proc_stats, parent_stats)

    def _process_statedump_block(self, period_data, **kwargs):
        dev = kwargs['dev']
        diskname = kwargs['diskname']
        if dev not in period_data.disks:
            period_data.disks[dev] = DiskStats(dev, diskname)
        else:
            period_data.disks[dev].diskname = diskname

    def _process_create_fd(self, period_data, **kwargs):
        timestamp = kwargs['timestamp']
        parent_proc = kwargs['parent_proc']
        tid = parent_proc.tid
        fd = kwargs['fd']

        if tid not in period_data.tids:
            period_data.tids[tid] = ProcessIOStats.new_from_process(
                parent_proc)

        parent_stats = period_data.tids[tid]
        if fd not in parent_stats.fds:
            parent_stats.fds[fd] = []
        parent_stats.fds[fd].append(FDStats.new_from_fd(parent_proc.fds[fd],
                                                        timestamp))

    def _process_close_fd(self, period_data, **kwargs):
        timestamp = kwargs['timestamp']
        parent_proc = kwargs['parent_proc']
        tid = parent_proc.tid
        fd = kwargs['fd']

        if tid not in period_data.tids:
            return
        parent_stats = period_data.tids[tid]
        last_fd = parent_stats.get_fd(fd)
        if last_fd is None:
            return
        last_fd.close_ts = timestamp

    def _process_update_fd(self, period_data, **kwargs):
        timestamp = kwargs['timestamp']
        parent_proc = kwargs['parent_proc']
        tid = parent_proc.tid
        fd = kwargs['fd']

        if fd not in parent_proc.fds:
            return

        if fd not in period_data.tids[tid].fds:
            period_data.tids[tid].fds[fd] = []
            period_data.tids[tid].fds[fd].append(
                FDStats.new_from_fd(parent_proc.fds[fd], timestamp))

        new_filename = parent_proc.fds[fd].filename
        fd_list = period_data.tids[tid].fds[fd]
        fd_list[-1].filename = new_filename


class DiskStats():
    MINORBITS = 20
    MINORMASK = ((1 << MINORBITS) - 1)

    def __init__(self, dev, diskname=None):
        self.dev = dev
        if diskname is not None:
            self.diskname = diskname
        else:
            self.diskname = DiskStats._get_name_from_dev(dev)

        self.min_rq_duration = None
        self.max_rq_duration = None
        self.total_rq_sectors = 0
        self.total_rq_duration = 0
        self.rq_list = []

    @classmethod
    def new_from_disk(cls, disk):
        return cls(disk.dev, disk.diskname)

    @property
    def rq_count(self):
        return len(self.rq_list)

    def update_stats(self, req):
        if self.min_rq_duration is None or req.duration < self.min_rq_duration:
            self.min_rq_duration = req.duration
        if self.max_rq_duration is None or req.duration > self.max_rq_duration:
            self.max_rq_duration = req.duration

        self.total_rq_sectors += req.nr_sector
        self.total_rq_duration += req.duration
        self.rq_list.append(req)

    def reset(self):
        self.min_rq_duration = None
        self.max_rq_duration = None
        self.total_rq_sectors = 0
        self.total_rq_duration = 0
        self.rq_list = []

    @staticmethod
    def _get_name_from_dev(dev):
        # imported from include/linux/kdev_t.h
        major = dev >> DiskStats.MINORBITS
        minor = dev & DiskStats.MINORMASK

        return '(%d,%d)' % (major, minor)


class IfaceStats():
    def __init__(self, name):
        self.name = name
        self.recv_bytes = 0
        self.recv_packets = 0
        self.sent_bytes = 0
        self.sent_packets = 0

    def reset(self):
        self.recv_bytes = 0
        self.recv_packets = 0
        self.sent_bytes = 0
        self.sent_packets = 0


class ProcessIOStats(stats.Process):
    def __init__(self, pid, tid, comm):
        super().__init__(pid, tid, comm)
        self.disk_io = stats.IO()
        self.net_io = stats.IO()
        self.unk_io = stats.IO()
        self.block_io = stats.IO()
        # FDStats objects, indexed by fd (fileno)
        self.fds = {}
        self.rq_list = []

    @classmethod
    def new_from_process(cls, proc):
        return cls(proc.pid, proc.tid, proc.comm)

    # Total read/write does not account for block layer I/O
    @property
    def total_read(self):
        return self.disk_io.read + self.net_io.read + self.unk_io.read

    @property
    def total_write(self):
        return self.disk_io.write + self.net_io.write + self.unk_io.write

    def update_fd_stats(self, req):
        if req.errno is not None:
            return

        if req.fd is None or self.get_fd(req.fd) is None:
            return

        self.get_fd(req.fd).update_stats(req)
        if isinstance(req, sv.ReadWriteIORequest):
            if req.fd_in is not None:
                self.get_fd(req.fd_in).update_stats(req)

            if req.fd_out is not None:
                self.get_fd(req.fd_out).update_stats(req)

    def update_block_stats(self, req):
        self.rq_list.append(req)

        if req.operation is sv.IORequest.OP_READ:
            self.block_io.read += req.size
        elif req.operation is sv.IORequest.OP_WRITE:
            self.block_io.write += req.size

    def update_io_stats(self, req, fd_types):
        self.rq_list.append(req)

        if req.size is None or req.errno is not None:
            return

        if req.operation is sv.IORequest.OP_READ:
            self._update_read(req.returned_size, fd_types['fd'])
        elif req.operation is sv.IORequest.OP_WRITE:
            self._update_write(req.returned_size, fd_types['fd'])
        elif req.operation is sv.IORequest.OP_READ_WRITE:
            self._update_read(req.returned_size, fd_types['fd_in'])
            self._update_write(req.returned_size, fd_types['fd_out'])

    def _update_read(self, size, fd_type):
        if fd_type == sv.FDType.disk:
            self.disk_io.read += size
        elif fd_type == sv.FDType.net or fd_type == sv.FDType.maybe_net:
            self.net_io.read += size
        else:
            self.unk_io.read += size

    def _update_write(self, size, fd_type):
        if fd_type == sv.FDType.disk:
            self.disk_io.write += size
        elif fd_type == sv.FDType.net or fd_type == sv.FDType.maybe_net:
            self.net_io.write += size
        else:
            self.unk_io.write += size

    def _get_current_fd(self, fd):
        fd_stats = self.fds[fd][-1]
        if fd_stats.close_ts is not None:
            return None

        return fd_stats

    @staticmethod
    def _get_fd_by_timestamp(fd_list, timestamp):
        """Return the FDStats object whose lifetime contains timestamp.

        This method performs a recursive binary search on the given
        fd_list argument, and will find the FDStats object for which
        the timestamp is contained between its open_ts and close_ts
        attributes.

        Args:
            fd_list (list): list of FDStats object, sorted
            chronologically by open_ts.

            timestamp (int): timestamp in nanoseconds (ns) since unix
            epoch which should be contained in the FD's lifetime.

        Returns:
            The FDStats object whose lifetime contains the given
            timestamp, None if no such object exists.
        """
        list_size = len(fd_list)
        if list_size == 0:
            return None

        midpoint = list_size // 2
        fd_stats = fd_list[midpoint]

        # Handle case of currently open fd (i.e. no close_ts)
        if fd_stats.close_ts is None:
            if timestamp >= fd_stats.open_ts:
                return fd_stats
        else:
            if fd_stats.open_ts <= timestamp <= fd_stats.close_ts:
                return fd_stats
            else:
                if timestamp < fd_stats.open_ts:
                    return ProcessIOStats._get_fd_by_timestamp(
                        fd_list[:midpoint], timestamp)
                else:
                    return ProcessIOStats._get_fd_by_timestamp(
                        fd_list[midpoint + 1:], timestamp)

    def get_fd(self, fd, timestamp=None):
        if fd not in self.fds or not self.fds[fd]:
            return None

        if timestamp is None:
            fd_stats = self._get_current_fd(fd)
        else:
            fd_stats = ProcessIOStats._get_fd_by_timestamp(self.fds[fd],
                                                           timestamp)

        return fd_stats

    def reset(self):
        self.disk_io.reset()
        self.net_io.reset()
        self.unk_io.reset()
        self.block_io.reset()
        self.rq_list = []

        for fd in self.fds:
            fd_stats = self.get_fd(fd)
            if fd_stats is not None:
                fd_stats.reset()


class FDStats():
    def __init__(self, fd, filename, fd_type, cloexec, family, open_ts):
        self.fd = fd
        self.filename = filename
        self.fd_type = fd_type
        self.cloexec = cloexec
        self.family = family
        self.open_ts = open_ts
        self.close_ts = None
        self.io = stats.IO()
        # IO Requests that acted upon the FD
        self.rq_list = []

    @classmethod
    def new_from_fd(cls, fd, open_ts):
        return cls(fd.fd, fd.filename, fd.fd_type, fd.cloexec, fd.family,
                   open_ts)

    def update_stats(self, req):
        if req.operation is sv.IORequest.OP_READ:
            self.io.read += req.returned_size
        elif req.operation is sv.IORequest.OP_WRITE:
            self.io.write += req.returned_size
        elif req.operation is sv.IORequest.OP_READ_WRITE:
            if self.fd == req.fd_in:
                self.io.read += req.returned_size
            elif self.fd == req.fd_out:
                self.io.write += req.returned_size

        self.rq_list.append(req)

    def reset(self):
        self.io.reset()
        self.rq_list = []


class FileStats():
    GENERIC_NAMES = ['pipe', 'socket', 'anon_inode', 'unknown']

    def __init__(self, filename):
        self.filename = filename
        self.io = stats.IO()
        # Dict of file descriptors representing this file, indexed by
        # parent pid
        # FIXME this doesn't cover FD reuse cases
        self.fd_by_pid = {}

    def update_stats(self, fd_stats, proc_stats):
        self.io += fd_stats.io

        if proc_stats.pid is not None:
            pid = proc_stats.pid
        else:
            pid = proc_stats.tid

        if pid not in self.fd_by_pid:
            self.fd_by_pid[pid] = fd_stats.fd

    def reset(self):
        self.io.reset()

    @staticmethod
    def is_generic_name(filename):
        for generic_name in FileStats.GENERIC_NAMES:
            if filename.startswith(generic_name):
                return True

        return False



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .sched import SchedStateProvider
from .mem import MemStateProvider
from .irq import IrqStateProvider
from .syscalls import SyscallsStateProvider
from .io import IoStateProvider
from .statedump import StatedumpStateProvider
from .block import BlockStateProvider
from .net import NetStateProvider
from .sv import MemoryManagement


class State:
    def __init__(self):
        self.cpus = {}
        self.tids = {}
        self.disks = {}
        self.mm = MemoryManagement()
        self._notification_cbs = {}
        # State changes can be handled differently depending on
        # version of tracer used, so keep track of it.
        self._tracer_version = None

    def register_notification_cbs(self, period_data, cbs):
        for name in cbs:
            if name not in self._notification_cbs:
                self._notification_cbs[name] = []
            # Store the callback in the form of (period_data, function)
            self._notification_cbs[name].append((period_data, cbs[name]))

    def send_notification_cb(self, name, **kwargs):
        if name in self._notification_cbs:
            for cb_tuple in self._notification_cbs[name]:
                cb_tuple[1](cb_tuple[0], **kwargs)

    def clear_period_notification_cbs(self, period_data):
        for name in self._notification_cbs:
            for cb in self._notification_cbs[name]:
                if cb[0] == period_data:
                    self._notification_cbs[name].remove(cb)


class Automaton:
    def __init__(self):
        self._state = State()
        self._state_providers = [
            SchedStateProvider(self._state),
            MemStateProvider(self._state),
            IrqStateProvider(self._state),
            SyscallsStateProvider(self._state),
            IoStateProvider(self._state),
            StatedumpStateProvider(self._state),
            BlockStateProvider(self._state),
            NetStateProvider(self._state)
        ]

    def process_event(self, ev):
        for sp in self._state_providers:
            sp.process_event(ev)

    @property
    def state(self):
        return self._state



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#               2015 - Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import errno
import operator
import statistics
from . import mi
from ..core import syscalls
from .command import Command


class SyscallsAnalysis(Command):
    _DESC = """The syscallstats command."""
    _ANALYSIS_CLASS = syscalls.SyscallsAnalysis
    _MI_TITLE = 'System call statistics'
    _MI_DESCRIPTION = 'Per-TID and global system call statistics'
    _MI_TAGS = [mi.Tags.SYSCALL, mi.Tags.STATS]
    _MI_TABLE_CLASS_PER_TID_STATS = 'per-tid'
    _MI_TABLE_CLASS_TOTAL = 'total'
    _MI_TABLE_CLASS_SUMMARY = 'summary'
    _MI_TABLE_CLASSES = [
        (
            _MI_TABLE_CLASS_PER_TID_STATS,
            'System call statistics', [
                ('syscall', 'System call', mi.Syscall),
                ('count', 'Call count', mi.Number, 'calls'),
                ('min_duration', 'Minimum call duration', mi.Duration),
                ('avg_duration', 'Average call duration', mi.Duration),
                ('max_duration', 'Maximum call duration', mi.Duration),
                ('stdev_duration', 'Call duration standard deviation',
                 mi.Duration),
                ('return_values', 'Return values count', mi.String),
            ]
        ),
        (
            _MI_TABLE_CLASS_TOTAL,
            'Per-TID system call statistics', [
                ('process', 'Process', mi.Process),
                ('count', 'Total system call count', mi.Number, 'calls'),
            ]
        ),
        (
            _MI_TABLE_CLASS_SUMMARY,
            'System call statistics - summary', [
                ('time_range', 'Time range', mi.TimeRange),
                ('process', 'Process', mi.Process),
                ('count', 'Total system call count', mi.Number, 'calls'),
            ]
        ),
    ]

    def _analysis_tick(self, period_data, end_ns):
        if period_data is None:
            return

        begin_ns = period_data.period.begin_evt.timestamp
        total_table, per_tid_tables = self._get_result_tables(period_data,
                                                              begin_ns, end_ns)

        if self._mi_mode:
            self._mi_append_result_tables(per_tid_tables)
            self._mi_append_result_table(total_table)
        else:
            self._print_date(begin_ns, end_ns)
            self._print_results(total_table, per_tid_tables)

    def _post_analysis(self):
        if not self._mi_mode:
            return

        if len(self._mi_get_result_tables(self._MI_TABLE_CLASS_TOTAL)) > 1:
            self._create_summary_result_table()

        self._mi_print()

    def _create_summary_result_table(self):
        total_tables = self._mi_get_result_tables(self._MI_TABLE_CLASS_TOTAL)
        begin = total_tables[0].timerange.begin.value
        end = total_tables[-1].timerange.end.value
        summary_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_SUMMARY,
                                         begin, end)

        for total_table in total_tables:
            for row in total_table.rows:
                process = row.process
                count = row.count
                summary_table.append_row(
                    time_range=total_table.timerange,
                    process=process,
                    count=count,
                )

        self._mi_clear_result_tables()
        self._mi_append_result_table(summary_table)

    def _get_result_tables(self, period_data, begin_ns, end_ns):
        per_tid_tables = []
        total_table = self._mi_create_result_table(self._MI_TABLE_CLASS_TOTAL,
                                                   begin_ns, end_ns)
        for proc_stats in sorted(period_data.tids.values(),
                                 key=operator.attrgetter('total_syscalls'),
                                 reverse=True):
            if proc_stats.total_syscalls == 0:
                continue

            pid = proc_stats.pid

            if proc_stats.pid is None:
                pid = '?'

            subtitle = '%s (%s, TID: %d)' % (proc_stats.comm, pid,
                                             proc_stats.tid)
            result_table = \
                self._mi_create_result_table(
                    self._MI_TABLE_CLASS_PER_TID_STATS, begin_ns, end_ns,
                    subtitle)

            for syscall in sorted(proc_stats.syscalls.values(),
                                  key=operator.attrgetter('count'),
                                  reverse=True):
                durations = []
                return_count = {}

                for syscall_event in syscall.syscalls_list:
                    durations.append(syscall_event.duration)

                    if syscall_event.ret >= 0:
                        return_key = 'success'
                    else:
                        try:
                            return_key = errno.errorcode[-syscall_event.ret]
                        except KeyError:
                            return_key = str(syscall_event.ret)

                    if return_key not in return_count:
                        return_count[return_key] = 1

                    return_count[return_key] += 1

                if len(durations) > 2:
                    stdev = mi.Duration(statistics.stdev(durations))
                else:
                    stdev = mi.Unknown()

                result_table.append_row(
                    syscall=mi.Syscall(syscall.name),
                    count=mi.Number(syscall.count),
                    min_duration=mi.Duration(syscall.min_duration),
                    avg_duration=mi.Duration(syscall.total_duration /
                                             syscall.count),
                    max_duration=mi.Duration(syscall.max_duration),
                    stdev_duration=stdev,
                    return_values=mi.String(str(return_count)),
                )

            per_tid_tables.append(result_table)
            total_table.append_row(
                process=mi.Process(proc_stats.comm, pid=proc_stats.pid,
                                   tid=proc_stats.tid),
                count=mi.Number(proc_stats.total_syscalls),
            )

        return total_table, per_tid_tables

    def _print_results(self, total_table, per_tid_tables):
        line_format = '{:<38} {:>14} {:>14} {:>14} {:>12} {:>10}  {:<14}'

        print('Per-TID syscalls statistics (usec)')
        total_calls = 0

        for total_row, table in zip(total_table.rows, per_tid_tables):
            print(line_format.format(table.subtitle,
                                     'Count', 'Min', 'Average', 'Max',
                                     'Stdev', 'Return values'))
            for row in table.rows:
                syscall_name = row.syscall.name
                syscall_count = row.count.value
                min_duration = round(row.min_duration.to_us(), 3)
                avg_duration = round(row.avg_duration.to_us(), 3)
                max_duration = round(row.max_duration.to_us(), 3)

                if type(row.stdev_duration) is mi.Unknown:
                    stdev = '?'
                else:
                    stdev = round(row.stdev_duration.to_us(), 3)

                proc_total_calls = total_row.count.value
                print(line_format.format(
                    ' - ' + syscall_name, syscall_count, min_duration,
                    avg_duration, max_duration, stdev,
                    row.return_values.value))

            print(line_format.format('Total:', proc_total_calls,
                                     '', '', '', '', ''))
            print('-' * 113)
            total_calls += proc_total_calls

        print('\nTotal syscalls: %d' % (total_calls))

    def _add_arguments(self, ap):
        Command._add_proc_filter_args(ap)


def _run(mi_mode):
    syscallscmd = SyscallsAnalysis(mi_mode=mi_mode)
    syscallscmd.run()


# entry point (human)
def run():
    _run(mi_mode=False)


# entry point (MI)
def run_mi():
    _run(mi_mode=True)



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class StateProvider:
    def __init__(self, state, cbs):
        self._state = state
        self._cbs = cbs

    def process_event(self, ev):
        name = ev.name

        if name in self._cbs:
            self._cbs[name](ev)
        # for now we process all the syscalls at the same place
        elif 'syscall_entry' in self._cbs and \
                (name.startswith('sys_') or name.startswith('syscall_entry_')):
            self._cbs['syscall_entry'](ev)
        elif 'syscall_exit' in self._cbs and \
                (name.startswith('exit_syscall') or
                 name.startswith('syscall_exit_')):
            self._cbs['syscall_exit'](ev)



# The MIT License (MIT)
#
# Copyright (C) 2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import stats
from .analysis import Analysis, PeriodData


class _PeriodData(PeriodData):
    def __init__(self):
        self.tids = {}
        self.total_syscalls = 0


class SyscallsAnalysis(Analysis):
    def __init__(self, state, conf):
        notification_cbs = {
            'syscall_exit': self._process_syscall_exit
        }
        super().__init__(state, conf, notification_cbs)

    def _create_period_data(self):
        return _PeriodData()

    def _process_syscall_exit(self, period_data, **kwargs):
        cpu_id = kwargs['cpu_id']
        proc = kwargs['proc']
        tid = proc.tid
        current_syscall = proc.current_syscall
        name = current_syscall.name

        if not self._filter_process(proc):
            return
        if not self._filter_cpu(cpu_id):
            return

        if tid not in period_data.tids:
            period_data.tids[tid] = ProcessSyscallStats.new_from_process(proc)

        proc_stats = period_data.tids[tid]
        if name not in proc_stats.syscalls:
            proc_stats.syscalls[name] = SyscallStats(name)

        proc_stats.syscalls[name].update_stats(current_syscall)
        proc_stats.total_syscalls += 1
        period_data.total_syscalls += 1


class ProcessSyscallStats(stats.Process):
    def __init__(self, pid, tid, comm):
        super().__init__(pid, tid, comm)

        # indexed by syscall name
        self.syscalls = {}
        self.total_syscalls = 0

    def reset(self):
        pass


class SyscallStats():
    def __init__(self, name):
        self.name = name
        self.min_duration = None
        self.max_duration = None
        self.total_duration = 0
        self.syscalls_list = []

    @property
    def count(self):
        return len(self.syscalls_list)

    def update_stats(self, syscall):
        duration = syscall.duration

        if self.min_duration is None or self.min_duration > duration:
            self.min_duration = duration
        if self.max_duration is None or self.max_duration < duration:
            self.max_duration = duration

        self.total_duration += duration
        self.syscalls_list.append(syscall)



# The MIT License (MIT)
#
# Copyright (C) 2016 - Julien Desfossez <jdesfossez@efficios.com>
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import math
import operator
import statistics
import collections
import ast
import re
from collections import OrderedDict
from . import mi, termgraph
from ..core import periods
from .command import Command


class _StatsFreqTables():
    def __init__(self):
        # Stats tables
        self.per_parent_stats_table = None
        self.per_parent_count_table = None
        self.per_parent_pc_table = None
        self.global_duration_table = None
        self.global_count_table = None
        self.global_pc_table = None

        # Raw values for the frequency distributions
        # *_values[period][child] = []
        self.duration_values = {}
        self.count_values = {}
        self.pc_values = {}
        self.global_duration_values = {}
        self.global_count_values = {}
        self.global_pc_values = {}

        # Freq tables
        self.per_parent_freq_tables = []
        self.per_parent_count_freq_tables = []
        self.per_parent_pc_freq_tables = []
        self.global_duration_freq_tables = []
        self.global_count_freq_tables = []
        self.global_pc_freq_tables = []


class _PeriodStats():
    def __init__(self, count=0, min=None, max=0, stdev=0, total=0):
        self.count = count
        self.min = min
        self.max = max
        self.stdev = stdev
        self.total = total
        self.count_array = []
        self.durations = []
        self.min_count = None
        self.max_count = 0
        self.total_count = 0
        # Percentage of the parent period time spent
        self.min_pc = None
        self.max_pc = 0
        self.total_pc = 0
        self.pc_array = []
        # How many parent periods have us as a child, indexed by
        # parent period name.
        self.parent_count = {}

    def add_count(self, count):
        if self.min_count is None or count < self.min_count:
            self.min_count = count
        if self.max_count < count:
            self.max_count = count
        self.total_count += count
        self.count_array.append(count)

    def add_duration(self, duration):
        if self.min is None or duration < self.min:
            self.min = duration
        if self.max < duration:
            self.max = duration
        self.total += duration
        self.durations.append(duration)

    def add_percentage(self, pc):
        if self.min_pc is None or pc < self.min_pc:
            self.min_pc = pc
        if self.max_pc < pc:
            self.max_pc = pc
        self.total_pc += pc
        self.pc_array.append(pc)


class _TmpAggregation():
    def __init__(self, parent=None):
        # self._children[name] = [durations]
        self._children = {}
        self._parent = parent
        self.capture_groups = None

    @property
    def children(self):
        return self._children

    def add_child(self, name, duration):
        if name not in self._children.keys():
            self._children[name] = []
        self._children[name].append(duration)

        parent = self._parent
        while parent is not None:
            parent.add_child(name, duration)
            parent = parent._parent


class _AggregatedPeriodStats():
    def __init__(self, registry, name):
        self._reg = registry
        self._name = name
        self._children = OrderedDict()
        self._stats = _PeriodStats()
        self.nr_periods = 0
        self._init_children()

    def _recurs_find_children(self, period):
        for child in period.children:
            self._children[child.name] = _PeriodStats()
            self._recurs_find_children(child)

    def _init_children(self):
        period_def = self._reg.get_period_def(self._name)
        if period_def is None:
            return
        self._recurs_find_children(period_def)

    def finish_period(self, start_ts, end_ts, child_dict):
        parent_duration = end_ts - start_ts
        for child in child_dict.keys():
            count = len(child_dict[child])
            duration = 0
            for period in child_dict[child]:
                duration += period
            c = self._children[child]
            pc = (duration / parent_duration) * 100

            c.add_count(count)
            c.add_duration(duration)
            c.add_percentage(pc)

            if self._name not in c.parent_count.keys():
                c.parent_count[self._name] = 0
            c.parent_count[self._name] += 1
        self.nr_periods += 1


class _AggregatedItem():
    def __init__(self, event, parent_event, group_by_captures, full_captures):
        self._event = event
        self._parent = parent_event
        self._group_by_captures = group_by_captures
        self._full_captures = full_captures

    @property
    def event(self):
        return self._event

    @property
    def parent_event(self):
        return self._parent

    @property
    def group_by_captures(self):
        return self._group_by_captures

    @property
    def full_captures(self):
        return self._full_captures


class PeriodAnalysisCommand(Command):
    _DESC = """The periods command."""
    _ANALYSIS_CLASS = periods.PeriodAnalysis
    _MI_TITLE = 'Periods analysis'
    _MI_DESCRIPTION = \
        'Periods frequency distribution, statistics, top, and log'
    _MI_TAGS = [mi.Tags.PERIOD, mi.Tags.STATS, mi.Tags.FREQ, mi.Tags.TOP,
                mi.Tags.LOG]
    _MI_TABLE_CLASS_LOG = 'log'
    _MI_TABLE_CLASS_TOP = 'top'
    _MI_TABLE_CLASS_PER_PERIOD_STATS = 'per_period_stats'
    _MI_TABLE_CLASS_PER_PARENT_STATS = 'per_parent_stats'
    _MI_TABLE_CLASS_PER_PARENT_COUNT = 'per_parent_count'
    _MI_TABLE_CLASS_PER_PARENT_PC = 'per_parent_percentage'
    _MI_TABLE_CLASS_FREQ_DURATION = 'freq_duration'
    _MI_TABLE_CLASS_FREQ_COUNT = 'freq_count'
    _MI_TABLE_CLASS_FREQ_PC = 'freq_ratio'
    _MI_TABLE_CLASS_HIERARCHICAL_LOG = 'aggregated_log'
    _MI_TABLE_CLASS_AGGREGATED_TOP = 'aggregated_top'
    _MI_TABLE_CLASS_AGGREGATED_LOG = 'aggregated_stats'
    _MI_TABLE_CLASSES = [
        (
            _MI_TABLE_CLASS_LOG,
            'Period log', [
                ('begin_ts', 'Period begin timestamp', mi.Timestamp),
                ('end_ts', 'Period end timestamp', mi.Timestamp),
                ('duration', 'Period duration', mi.Duration),
                ('name', 'Period name', mi.String),
                ('begin_captures', 'Begin captures', mi.String),
                ('end_captures', 'End captures', mi.String),
            ]
        ),
        (
            _MI_TABLE_CLASS_TOP,
            'Period top', [
                ('begin_ts', 'Period begin timestamp', mi.Timestamp),
                ('end_ts', 'Period end timestamp', mi.Timestamp),
                ('duration', 'Period duration', mi.Duration),
                ('name', 'Period name', mi.String),
                ('begin_captures', 'Begin captures', mi.String),
                ('end_captures', 'End captures', mi.String),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_PERIOD_STATS,
            'Period statistics', [
                ('name', 'Period name', mi.String),
                ('count', 'Period count', mi.Number, 'occurences'),
                ('min_duration', 'Minimum duration', mi.Duration),
                ('avg_duration', 'Average duration', mi.Duration),
                ('max_duration', 'Maximum duration', mi.Duration),
                ('stdev_duration', 'Period duration standard deviation',
                 mi.Duration),
                ('runtime', 'Total runtime', mi.Duration),
            ]
        ),
        (
            _MI_TABLE_CLASS_FREQ_DURATION,
            'Period duration frequency distribution', [
                ('lower', 'Duration (lower bound)', mi.Duration),
                ('upper', 'Duration (upper bound)', mi.Duration),
                ('count', 'Period duration', mi.Number, 'us'),
            ]
        ),
        (
            _MI_TABLE_CLASS_FREQ_COUNT,
            'Period count frequency distribution', [
                ('lower', 'Count (lower bound)', mi.Number),
                ('upper', 'Count (upper bound)', mi.Number),
                ('count', 'Period count', mi.Number, 'count'),
            ]
        ),
        (
            _MI_TABLE_CLASS_FREQ_PC,
            'Period usage ratio frequency distribution', [
                ('lower', 'Ratio (lower bound)', mi.Number),
                ('upper', 'Ratio (upper bound)', mi.Number),
                ('count', 'Period usage ratio', mi.Number, '%'),
            ]
        ),
        (
            _MI_TABLE_CLASS_HIERARCHICAL_LOG,
            'Hierarchical period log', [
                ('parent_begin_ts', 'Parent begin timestamp', mi.Timestamp),
                ('parent_end_ts', 'Parent end timestamp', mi.Timestamp),
                ('parent_name', 'Parent period name', mi.String),
                ('child_begin_ts', 'Child begin timestamp', mi.Timestamp),
                ('child_end_ts', 'Child end timestamp', mi.Timestamp),
                ('child_name', 'Child period name', mi.String),
                ('child_duration', 'Child period duration', mi.Duration),
                ('parent_duration', 'Parent period duration', mi.Duration),
                ('captures', 'Captures', mi.String),
            ]
        ),
        (
            _MI_TABLE_CLASS_AGGREGATED_TOP,
            'Aggregated period top', [
                ('parent_begin_ts', 'Parent begin timestamp', mi.Timestamp),
                ('parent_end_ts', 'Parent end timestamp', mi.Timestamp),
                ('parent_name', 'Parent period name', mi.String),
                ('child_begin_ts', 'Child begin timestamp', mi.Timestamp),
                ('child_end_ts', 'Child end timestamp', mi.Timestamp),
                ('child_name', 'Child period name', mi.String),
                ('child_duration', 'Child period duration', mi.Duration),
                ('parent_duration', 'Parent period duration', mi.Duration),
                ('captures', 'Captures', mi.String),
            ]
        ),
        (
            _MI_TABLE_CLASS_AGGREGATED_LOG,
            'Aggregated log', [
                ('parent_name', 'Parent period name', mi.String),
                ('parent_begin_ts', 'Parent begin timestamp', mi.Timestamp),
                ('parent_end_ts', 'Parent end timestamp', mi.Timestamp),
                ('child_name', 'Child period name', mi.String),
                ('count', 'Period count', mi.Number, 'occurences'),
                ('min_duration', 'Minimum duration', mi.Duration),
                ('avg_duration', 'Average duration', mi.Duration),
                ('max_duration', 'Maximum duration', mi.Duration),
                ('stdev_duration', 'Period duration standard deviation',
                 mi.Duration),
                ('runtime', 'Total runtime', mi.Duration),
                ('parent_captures', 'Parent captures', mi.String),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_PARENT_STATS,
            'Per-parent period duration statistics', [
                ('name', 'Period name', mi.String),
                ('parent', 'Parent', mi.String),
                ('min_duration', 'Minimum duration', mi.Duration),
                ('avg_duration', 'Average duration', mi.Duration),
                ('max_duration', 'Maximum duration', mi.Duration),
                ('stdev_duration', 'Period duration standard deviation',
                 mi.Duration),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_PARENT_COUNT,
            'Per-parent period count statistics', [
                ('name', 'Period name', mi.String),
                ('parent', 'Parent', mi.String),
                ('min', 'Minimum', mi.Number, 'occurences'),
                ('avg', 'Average', mi.Number, 'occurences'),
                ('max', 'Maximum', mi.Number, 'occurences'),
                ('stdev', 'Standard deviation', mi.Duration),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_PARENT_PC,
            'Per-parent duration ratio', [
                ('name', 'Period name', mi.String),
                ('parent', 'Parent', mi.String),
                ('min', 'Minimum', mi.Number, 'occurences'),
                ('avg', 'Average', mi.Number, 'occurences'),
                ('max', 'Maximum', mi.Number, 'occurences'),
                ('stdev', 'Standard deviation', mi.Duration),
            ]
        ),
    ]

    def _filter_duration(self, duration):
        if self._args.min_duration is not None and \
                duration < (self._args.min_duration * 1000):
            return False
        if self._args.max_duration is not None and \
                duration > (self._args.max_duration * 1000):
            return False
        return True

    def _filter_event_duration(self, period_event):
        return self._filter_duration(period_event.duration)

    def _get_period_tree(self, period, period_tree):
        period_tree[period.name] = OrderedDict()
        for child in period.children:
            self._get_period_tree(child, period_tree[period.name])

    def _analysis_tick(self, period_data, end_ns):
        # We only output something at the end of the analysis
        # not when each period finishes
        if period_data is not None:
            return

        # Override the timestamps since we are only interested in the
        # whole analysis timestamps, not the ones from the last period.
        begin_ns = self._analysis.first_event_ts
        end_ns = self._analysis.last_event_ts
        log_table = None
        top_table = None
        per_period_stats_table = None
        per_period_freq_tables = None

        aggregated_groups = None
        hierarchical_list = None
        aggregated_log_tables = None
        per_parent_aggregated_dict = None

        per_parent_stats_freq_group_by_tables = OrderedDict()
        per_period_stats_group_by_tables = OrderedDict()
        per_period_freq_group_by_tables = OrderedDict()
        # for freq-series:
        # freq_tables_group_per_period_names[group][period] = freq_table
        freq_tables_group_per_period_names = OrderedDict()

        # First pass to find the uniform values if needed
        freq_min = freq_max = freq_step = None

        period_tree = OrderedDict()
        reg = self._analysis_conf.period_def_registry
        for parent in reg.root_period_defs:
            self._get_period_tree(parent, period_tree)

        if self._args.select or self._args.order_by == "hierarchy" or \
                self._args.stats or self._args.freq:
            per_parent_aggregated_dict, hierarchical_list, per_period_stats, \
                per_parent_period_group_by_stats, \
                per_period_group_by_stats = self._get_aggregated_lists()
            if self._args.group_by:
                aggregated_groups = self._get_aggregated_groups(
                    per_parent_aggregated_dict)

        if self._args.log:
            # hierarchical view
            if self._args.order_by == "hierarchy":
                log_table = self._get_log_result_table(
                    begin_ns, end_ns, hierarchical_list)
            # aggregated view
            elif self._args.select:
                aggregated_log_tables = \
                    self._get_aggregated_log_table(
                        begin_ns, end_ns,
                        per_parent_aggregated_dict, aggregated_groups,
                        top=True)
            else:
                # time-based view
                log_table = self._get_log_result_table(
                    begin_ns, end_ns, self._analysis.all_period_list)

        if self._args.top:
            top_table = self._get_top_result_table(
                begin_ns, end_ns, self._analysis.all_period_list)

        # Common tables for stats and freq
        if self._args.stats or self._args.freq:
            per_period_stats_table = \
                self._get_per_period_stats_result_table(begin_ns, end_ns,
                                                        period_tree)
            per_parent_stats_freq_tables = \
                self._get_per_parent_stats_result_table(
                    begin_ns, end_ns, per_period_stats,
                    '', per_period_stats)

            if self._args.freq_uniform:
                for group in per_parent_period_group_by_stats.keys():
                    freq_min, freq_max, freq_step = \
                        self._find_filtered_uniform_freq_values(
                            per_period_group_by_stats[group])

            for group in per_parent_period_group_by_stats.keys():
                per_period_stats_group_by_tables[group], \
                    per_period_freq_group_by_tables[group], \
                    freq_tables_group_per_period_names[group] = \
                    self._get_grouped_by_period_stats_freq(
                        begin_ns, end_ns,
                        per_period_group_by_stats[group],
                        "'%s' - " % group,
                        freq_min, freq_max, freq_step)
                # One _StatsFreqTables per group
                per_parent_stats_freq_group_by_tables[group] = \
                    self._get_per_parent_stats_result_table(
                        begin_ns, end_ns,
                        per_parent_period_group_by_stats[group],
                        "'%s' - " % group, per_period_stats)

        if self._args.freq:
            per_period_freq_tables = \
                self._get_per_period_freq_result_tables(begin_ns, end_ns)
            # This updates per_parent_stats_freq_tables with the new tables,
            # nothing to return.
            self._get_per_parent_freq_result_table(
                begin_ns, end_ns, per_parent_stats_freq_tables)

            for group in per_parent_period_group_by_stats.keys():
                self._get_per_parent_freq_result_table(
                    begin_ns, end_ns,
                    per_parent_stats_freq_group_by_tables[group],
                    "'%s' -- " % group)

        if self._mi_mode:
            if log_table:
                self._mi_append_result_table(log_table)

            if top_table:
                self._mi_append_result_table(top_table)

            if self._args.stats:
                self._mi_append_result_table(per_period_stats_table)
                self._mi_append_result_table(
                    per_parent_stats_freq_tables.per_parent_stats_table)
                self._mi_append_result_table(
                    per_parent_stats_freq_tables.per_parent_count_table)
                self._mi_append_result_table(
                    per_parent_stats_freq_tables.per_parent_pc_table)
                self._mi_append_result_table(
                    per_parent_stats_freq_tables.global_duration_table)
                self._mi_append_result_table(
                    per_parent_stats_freq_tables.global_count_table)
                self._mi_append_result_table(
                    per_parent_stats_freq_tables.global_pc_table)

                for group in per_parent_period_group_by_stats.keys():
                    self._mi_append_result_table(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_stats_table)
                    self._mi_append_result_table(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_count_table)
                    self._mi_append_result_table(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_pc_table)
                    self._mi_append_result_table(
                        per_parent_stats_freq_group_by_tables[group].
                        global_duration_table)
                    self._mi_append_result_table(
                        per_parent_stats_freq_group_by_tables[group].
                        global_count_table)
                    self._mi_append_result_table(
                        per_parent_stats_freq_group_by_tables[group].
                        global_pc_table)

            if self._args.freq:
                self._mi_append_result_tables(per_period_freq_tables)
                self._mi_append_result_tables(
                    per_parent_stats_freq_tables.per_parent_freq_tables)
                self._mi_append_result_tables(
                    per_parent_stats_freq_tables.per_parent_count_freq_tables)
                self._mi_append_result_tables(
                    per_parent_stats_freq_tables.per_parent_pc_freq_tables)
                self._mi_append_result_tables(
                    per_parent_stats_freq_tables.global_duration_freq_tables)
                self._mi_append_result_tables(
                    per_parent_stats_freq_tables.global_count_freq_tables)
                self._mi_append_result_tables(
                    per_parent_stats_freq_tables.global_pc_freq_tables)

                for group in per_parent_period_group_by_stats.keys():
                    self._mi_append_result_tables(
                        per_period_freq_group_by_tables[group])
                    self._mi_append_result_tables(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_freq_tables)
                    self._mi_append_result_tables(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_count_freq_tables)
                    self._mi_append_result_tables(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_pc_freq_tables)
                    self._mi_append_result_tables(
                        per_parent_stats_freq_group_by_tables[group].
                        global_duration_freq_tables)
                    self._mi_append_result_tables(
                        per_parent_stats_freq_group_by_tables[group].
                        global_count_freq_tables)
                    self._mi_append_result_tables(
                        per_parent_stats_freq_group_by_tables[group].
                        global_pc_freq_tables)

                if self._args.freq_series:
                    per_period_tables_group_series = \
                        self._get_per_group_freq_series_tables(
                            begin_ns, end_ns, per_period_freq_group_by_tables,
                            freq_tables_group_per_period_names)
                    for period in per_period_tables_group_series.keys():
                        self._mi_append_result_tables(
                            [per_period_tables_group_series[period]])

        else:
            self._print_date(begin_ns, end_ns)

            if self._args.stats:
                self._print_per_period_stats(per_period_stats_table,
                                             period_tree)
                self._print_per_parent_stats(
                    per_parent_stats_freq_tables.per_parent_stats_table)
                self._print_per_parent_pc(
                    per_parent_stats_freq_tables.per_parent_pc_table)
                self._print_per_parent_count(
                    per_parent_stats_freq_tables.per_parent_count_table)
                self._print_per_parent_stats(
                    per_parent_stats_freq_tables.global_duration_table)
                self._print_per_parent_pc(
                    per_parent_stats_freq_tables.global_pc_table)
                self._print_per_parent_count(
                    per_parent_stats_freq_tables.global_count_table)

                for group in per_parent_period_group_by_stats.keys():
                    print("\n\n### Group: %s ###" % group)
                    self._print_per_period_stats(
                        per_period_stats_group_by_tables[group],
                        period_tree)
                    self._print_per_parent_stats(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_stats_table)
                    self._print_per_parent_pc(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_pc_table)
                    self._print_per_parent_count(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_count_table)
                    self._print_per_parent_stats(
                        per_parent_stats_freq_group_by_tables[group].
                        global_duration_table)
                    self._print_per_parent_pc(
                        per_parent_stats_freq_group_by_tables[group].
                        global_pc_table)
                    self._print_per_parent_count(
                        per_parent_stats_freq_group_by_tables[group].
                        global_count_table)

            if self._args.freq:
                self._print_freq(per_period_freq_tables, 'us')
                self._print_freq(
                    per_parent_stats_freq_tables.per_parent_freq_tables, 'us')
                self._print_freq(
                    per_parent_stats_freq_tables.per_parent_pc_freq_tables,
                    '%')
                self._print_freq(
                    per_parent_stats_freq_tables.per_parent_count_freq_tables,
                    'instances')
                self._print_freq(
                    per_parent_stats_freq_tables.global_duration_freq_tables,
                    'us')
                self._print_freq(
                    per_parent_stats_freq_tables.global_pc_freq_tables, '%')
                self._print_freq(
                    per_parent_stats_freq_tables.global_count_freq_tables,
                    'instances')

                for group in per_parent_period_group_by_stats.keys():
                    print("\n\n### Group: %s ###" % group)
                    self._print_freq(per_period_freq_group_by_tables[group],
                                     'us')
                    self._print_freq(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_freq_tables, 'us')
                    self._print_freq(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_pc_freq_tables,
                        '%')
                    self._print_freq(
                        per_parent_stats_freq_group_by_tables[group].
                        per_parent_count_freq_tables,
                        'instances')
                    self._print_freq(
                        per_parent_stats_freq_group_by_tables[group].
                        global_duration_freq_tables,
                        'us')
                    self._print_freq(
                        per_parent_stats_freq_group_by_tables[group].
                        global_pc_freq_tables, '%')
                    self._print_freq(
                        per_parent_stats_freq_group_by_tables[group].
                        global_count_freq_tables,
                        'instances')

            if log_table:
                self._print_period_events(log_table)

            if top_table:
                self._print_period_events(top_table)

            if aggregated_log_tables:
                self._print_aggregated_log(aggregated_log_tables)

    def _get_filtered_min_max_count_avg_total_values(self, durations):
        min = None
        max = None
        count = 0
        avg = 0
        total = 0
        filter_list = []
        for d in durations:
            if not self._filter_duration(d):
                continue
            if min is None or min > d:
                min = d
            if max is None or max < d:
                max = d
            count += 1
            total += d
            filter_list.append(d)
        if count > 0:
            avg = total / count
        else:
            avg = 0
        return min, max, count, avg, total, filter_list

    def _get_filtered_min_max_count_avg_total_flist(self, period_list):
        min = None
        max = None
        count = 0
        avg = 0
        total = 0
        filter_list = []
        for period_event in period_list:
            if not self._filter_event_duration(period_event):
                continue
            if min is None or min > period_event.duration:
                min = period_event.duration
            if max is None or max < period_event.duration:
                max = period_event.duration
            count += 1
            total += period_event.duration
            filter_list.append(period_event)
        if count > 0:
            avg = total / count
        else:
            avg = 0
        return min, max, count, avg, total, filter_list

    def _get_agg_filtered_min_max_count_avg_total_flist(self, ag_list):
        min = None
        max = None
        count = 0
        avg = 0
        total = 0
        filter_list = []
        for ag_event in ag_list:
            period_event = ag_event.event
            if not self._filter_event_duration(period_event):
                continue
            if min is None or min > period_event.duration:
                min = period_event.duration
            if max is None or max < period_event.duration:
                max = period_event.duration
            count += 1
            total += period_event.duration
            filter_list.append(period_event)
        if count > 0:
            avg = total / count
        else:
            avg = 0
        return min, max, count, avg, total, filter_list

    def _find_aggregated_subperiods(self, root, event, aggregated_list,
                                    group_by_captures,
                                    full_captures):
        if len(self._analysis_conf._select) == 0 or \
                event.name in self._analysis_conf._select:
            aggregated_list.append(_AggregatedItem(event, root,
                                                   group_by_captures,
                                                   full_captures))
        for capture in event.filtered_captures(
                self._analysis_conf._group_by):
            group_by_captures.append(capture)
        for capture in event.full_captures():
            full_captures.append(capture)
        for child in event.children:
            self._find_aggregated_subperiods(root, child, aggregated_list,
                                             group_by_captures,
                                             full_captures)

    def _add_parent_per_group_active_periods(self, event,
                                             per_group_active_periods,
                                             group_key):
        p = None
        if event.parent is not None and \
                event.parent not in per_group_active_periods[group_key].keys():
            p = self._add_parent_per_group_active_periods(
                event.parent, per_group_active_periods, group_key)
        per_group_active_periods[group_key][event] = _TmpAggregation(p)
        return per_group_active_periods[group_key][event]

    def _account_parents_in_group(self, event, full_captures,
                                  per_parent_period_group_by_stats,
                                  per_group_active_periods,
                                  per_period_group_by_stats):
        for g in full_captures:
            group_key = ''
            if len(g) < len(self._analysis_conf._group_by.keys()):
                continue
            for group in sorted(g, key=lambda x: x[0]):
                if len(group_key) == 0:
                    group_key = '%s = %s' % (group[0], group[1])
                else:
                    group_key = '%s, %s = %s' % (group_key, group[0],
                                                 group[1])

            if len(group_key) == 0:
                continue

            if group_key not in per_group_active_periods.keys():
                per_group_active_periods[group_key] = OrderedDict()

            # Statistics for this event alone in this group
            if group_key not in per_period_group_by_stats.keys():
                per_period_group_by_stats[group_key] = OrderedDict()
            if event.name not in per_period_group_by_stats[group_key].keys():
                per_period_group_by_stats[group_key][event.name] = \
                    _PeriodStats()
            per_period_group_by_stats[group_key][event.name].add_duration(
                event.duration)

            if group_key not in per_parent_period_group_by_stats.keys():
                per_parent_period_group_by_stats[group_key] = OrderedDict()
            if event.name not in \
                    per_parent_period_group_by_stats[group_key].keys():
                per_parent_period_group_by_stats[group_key][event.name] = \
                    _AggregatedPeriodStats(
                        self._analysis_conf.period_def_registry,
                        event.name)
            # Account all parent periods of this event in all of its groups
            _parent = event.parent
            _child = event
            while _parent is not None:
                if _parent not in per_group_active_periods[group_key].keys():
                    self._add_parent_per_group_active_periods(
                        _parent, per_group_active_periods, group_key)
                if _parent.name not in \
                        per_parent_period_group_by_stats[group_key].keys():
                    per_parent_period_group_by_stats[group_key][_parent.name] \
                        = _AggregatedPeriodStats(
                        self._analysis_conf.period_def_registry,
                        _parent.name)
                per_group_active_periods[group_key][_parent].add_child(
                    _child.name, _child.duration)
                _parent = _parent.parent

            if event in per_group_active_periods[group_key].keys():
                per_parent_period_group_by_stats[group_key][event.name]. \
                    finish_period(
                        event.start_ts, event.end_ts,
                        per_group_active_periods[group_key][event].children)

    def _hierarchical_sub(self, tmp_hierarchical_list, event, per_period_stats,
                          per_parent_period_group_by_stats, active_periods,
                          ancestors_captures, per_group_active_periods,
                          per_period_group_by_stats):
        tmp_hierarchical_list.append(event)

        event_captures = event.filtered_captures(self._analysis_conf._group_by)
#        print(parent_captures, event_captures)

        # Our local level capture to return to our parent combined with the
        # captures of our children.
        local_captures = []
        global_captures = []

        # Recursively iterate over all the children of this period
        for child in event.children:
            if not self._filter_event_duration(child):
                continue
            if child.name not in per_period_stats.keys():
                per_period_stats[child.name] = _AggregatedPeriodStats(
                    self._analysis_conf.period_def_registry,
                    child.name)
            active_periods[event].add_child(child.name, child.duration)
            active_periods[child] = _TmpAggregation(active_periods[event])
            child_captures = self._hierarchical_sub(
                tmp_hierarchical_list, child, per_period_stats,
                per_parent_period_group_by_stats, active_periods,
                ancestors_captures + event_captures,
                per_group_active_periods, per_period_group_by_stats)
            del(active_periods[child])
            for c in child_captures:
                local_captures.append(event_captures + c)
                global_captures.append(event_captures.copy() + c)
        if len(local_captures) == 0:
            local_captures = [event_captures]
            global_captures = [event_captures.copy()]
        full_captures = []
        for c in global_captures:
            tmp_c = c.copy()
            for d in ancestors_captures:
                tmp_c.append(d)
            # dedup
            if tmp_c not in full_captures:
                full_captures.append(tmp_c)
        active_periods[event].capture_groups = full_captures

        self._account_parents_in_group(event, full_captures,
                                       per_parent_period_group_by_stats,
                                       per_group_active_periods,
                                       per_period_group_by_stats)

        per_period_stats[event.name].finish_period(
            event.start_ts, event.end_ts,
            active_periods[event].children)

        return local_captures

    def _get_aggregated_lists(self):
        # Dict with parent period as key. Each entry contains a dict
        # of all child period that each contain a list of _AggregatedItem.
        # parent_aggregated_dict[parent_period][child_period] = []
        parent_aggregated_dict = {}
        # List of PeriodEvent ordered in hierarchy (parents are followed
        # by their children)
        hierarchical_list = []
        # dict of _AggregatedPeriodStats
        # OrderedDict because we want the same order as the period_tree
        per_period_stats = OrderedDict()
        per_parent_period_group_by_stats = OrderedDict()
        # Just the stats for the period per group (not relative to
        # its parents)
        per_period_group_by_stats = OrderedDict()
        # active_periods[period_event] = _TmpAggregation()
        active_periods = {}
        per_group_active_periods = {}
        for period_event in self._analysis.all_period_list:
            if not self._filter_event_duration(period_event):
                continue
            if self._analysis_conf._order_by == "hierarchy" or \
                    self._args.stats or self._args.freq:
                # Only top-level events to start the recursive iteration
                # and extract per_parent stats/freq and hierarchical list
                # of periods
                if period_event.parent is None:
                    active_periods[period_event] = _TmpAggregation()
                    if period_event.name not in per_period_stats.keys():
                        per_period_stats[period_event.name] = \
                            _AggregatedPeriodStats(
                                self._analysis_conf.period_def_registry,
                                period_event.name)

                    tmp_hierarchical_list = []
                    self._hierarchical_sub(
                        tmp_hierarchical_list, period_event, per_period_stats,
                        per_parent_period_group_by_stats, active_periods, [],
                        per_group_active_periods, per_period_group_by_stats)
                    del(active_periods[period_event])
                    for item in tmp_hierarchical_list:
                        hierarchical_list.append(item)

            if period_event.name != self._analysis_conf._aggregate_by:
                continue
            if period_event not in parent_aggregated_dict.keys():
                parent_aggregated_dict[period_event] = {}
            # Associate the periods with their full capture list (each period
            # sees its own capture and the capture of all its children)
            tmp_list = []
            for child in period_event.children:
                if not self._filter_event_duration(child):
                    continue
                self._find_aggregated_subperiods(
                    period_event,
                    child, tmp_list,
                    period_event.filtered_captures(
                        self._analysis_conf._group_by),
                    period_event.full_captures())
            for item in tmp_list:
                if item.event.name not in \
                        parent_aggregated_dict[period_event].keys():
                    parent_aggregated_dict[period_event][item.event.name] = []
                parent_aggregated_dict[period_event][item.event.name]. \
                    append(item)
        ordered_parent = collections.OrderedDict(
            sorted(parent_aggregated_dict.items(),
                   key=lambda t: t[0].start_ts))

        return ordered_parent, hierarchical_list, per_period_stats, \
            per_parent_period_group_by_stats, per_period_group_by_stats

    def _get_aggregated_groups(self, per_parent_aggregated_dict):
        # Group and flatten event list by captured keys, aggregate by parent
        # groups[group_key][parent][child] = [_AggregatedItem, ...]
        groups = {}
        for parent in per_parent_aggregated_dict.keys():
            for child in per_parent_aggregated_dict[parent].keys():
                for ag_event in per_parent_aggregated_dict[parent][child]:
                    group_key = ""
                    for group in sorted(ag_event.group_by_captures,
                                        key=lambda x: x[0]):
                        if len(group_key) == 0:
                            group_key = "%s = %s" % (group[0], group[1])
                        else:
                            group_key = "%s, %s = %s" % (group_key, group[0],
                                                         group[1])

                    if group_key not in groups.keys():
                        groups[group_key] = {}
                    if parent not in groups[group_key].keys():
                        groups[group_key][parent] = {}
                    if child not in groups[group_key][parent].keys():
                        groups[group_key][parent][child] = []
                    groups[group_key][parent][child].append(ag_event)
        return groups

    def _get_total_period_lists_stats(self):
        if self._args.min_duration is None and \
                self._args.max_duration is None:
            total_list = self._analysis.all_period_list
            stdev = self._compute_period_duration_stdev(total_list)
            total_stats = _PeriodStats(
                count=self._analysis.all_count,
                min=self._analysis.all_min_duration,
                max=self._analysis.all_max_duration,
                stdev=stdev,
                total=self._analysis.all_total_duration
            )
        else:
            min, max, count, avg, total, total_list = \
                self._get_filtered_min_max_count_avg_total_flist(
                    self._analysis.all_period_list)
            total_stats = _PeriodStats(
                count=count,
                min=min,
                max=max,
                stdev=self._compute_period_duration_stdev(total_list),
                total=total,
            )

        return [total_list], total_stats

    def _get_one_hierarchical_log_table(self, begin_ns, end_ns,
                                        aggregated_list, sub, top):
        if top:
            table = self._mi_create_result_table(
                self._MI_TABLE_CLASS_AGGREGATED_TOP, begin_ns, end_ns,
                subtitle=sub)
            top_events = sorted(aggregated_list, key=operator.attrgetter(
                'event.duration'), reverse=True)
            top_events = top_events[:self._args.limit]
            for ag_event in top_events:
                table.append_row(
                    parent_begin_ts=mi.Timestamp(
                        ag_event.parent_event.start_ts),
                    parent_end_ts=mi.Timestamp(
                        ag_event.parent_event.end_ts),
                    parent_name=mi.String(ag_event.parent_event.name),
                    child_begin_ts=mi.Timestamp(ag_event.event.start_ts),
                    child_end_ts=mi.Timestamp(ag_event.event.end_ts),
                    child_name=mi.String(ag_event.event.name),
                    child_duration=mi.Duration(ag_event.event.duration),
                    parent_duration=mi.Duration(
                        ag_event.parent_event.duration),
                    captures=mi.String(str(ag_event.full_captures)),
                )
        else:
            table = self._mi_create_result_table(
                self._MI_TABLE_CLASS_HIERARCHICAL_LOG, begin_ns, end_ns,
                subtitle=sub)
            for ag_event in aggregated_list:
                table.append_row(
                    parent_begin_ts=mi.Timestamp(
                        ag_event.parent_event.start_ts),
                    parent_end_ts=mi.Timestamp(ag_event.parent_event.end_ts),
                    parent_name=mi.String(ag_event.parent_event.name),
                    child_begin_ts=mi.Timestamp(ag_event.event.start_ts),
                    child_end_ts=mi.Timestamp(ag_event.event.end_ts),
                    child_name=mi.String(ag_event.event.name),
                    child_duration=mi.Duration(ag_event.event.duration),
                    parent_duration=mi.Duration(
                        ag_event.parent_event.duration),
                    captures=mi.String(str(ag_event.full_captures)),
                )
        return table

    def _get_hierarchical_log_top_result_table(
            self, begin_ns, end_ns, aggregated_list, aggregated_groups,
            top=False):
        result_tables = []
        ag_list = ""
        for i in self._analysis_conf._select:
            if len(ag_list) == 0:
                ag_list = i
            else:
                ag_list = "%s, %s" % (ag_list, i)
        sub = "Aggregation of (%s) by %s" % (
            ag_list, self._analysis_conf._aggregate_by)

        if aggregated_groups is None:
            table = self._get_one_hierarchical_log_table(begin_ns, end_ns,
                                                         aggregated_list, sub,
                                                         top)
            result_tables.append(table)
        else:
            for group in aggregated_groups.keys():
                group_sub = "%s, group: %s" % (sub, group)
                result_tables.append(self._get_one_hierarchical_log_table(
                    begin_ns, end_ns, aggregated_groups[group], group_sub,
                    top))

        return result_tables

    def _get_full_period_path(self, period_name):
        if len(period_name) == 0:
            return period_name
        return self._analysis_conf.period_def_registry.period_full_path(
            period_name)

    def _get_log_result_table(self, begin_ns, end_ns, period_list):
        result_table = self._mi_create_result_table(self._MI_TABLE_CLASS_LOG,
                                                    begin_ns, end_ns)
        for period_event in period_list:
            if not self._filter_event_duration(period_event):
                continue
            result_table.append_row(
                begin_ts=mi.Timestamp(period_event.start_ts),
                end_ts=mi.Timestamp(period_event.end_ts),
                duration=mi.Duration(period_event.duration),
                name=mi.String(self._get_full_period_path(period_event.name)),
                begin_captures=mi.String(period_event.begin_captures),
                end_captures=mi.String(period_event.end_captures),
            )
        return result_table

    def _get_top_result_table(self, begin_ns, end_ns, event_list):
        result_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_TOP, begin_ns, end_ns)

        top_events = sorted(event_list,
                            key=operator.attrgetter('duration'),
                            reverse=True)
        count = 0

        for period_event in top_events:
            if not self._filter_event_duration(period_event):
                continue
            if self._args.select and period_event.name not in \
                    self._args.select:
                continue
            result_table.append_row(
                begin_ts=mi.Timestamp(period_event.start_ts),
                end_ts=mi.Timestamp(period_event.end_ts),
                duration=mi.Duration(period_event.duration),
                name=mi.String(period_event.name),
                begin_captures=mi.String(period_event.begin_captures),
                end_captures=mi.String(period_event.end_captures),
            )
            count += 1
            if count == self._args.limit:
                break
        return result_table

    def _get_ordered_period_stats_list(self, parent_name, period_stats_list,
                                       period_tree):
        if parent_name not in self._analysis.all_period_stats.keys():
            return
        period_stats_list.append(self._analysis.all_period_stats[parent_name])
        for child in sorted(period_tree.keys()):
            self._get_ordered_period_stats_list(child, period_stats_list,
                                                period_tree[child])

    def _get_per_parent_stats_result_table(self, begin_ns, end_ns,
                                           per_period_stats, group_prefix,
                                           not_grouped_per_period_stats):
        duration_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_PARENT_STATS, begin_ns, end_ns,
            subtitle="%sWith active children" % group_prefix)
        count_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_PARENT_COUNT, begin_ns, end_ns,
            subtitle="%sWith active children" % group_prefix)
        global_duration_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_PARENT_STATS, begin_ns, end_ns,
            subtitle="%sGlobally" % group_prefix)
        global_count_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_PARENT_COUNT, begin_ns, end_ns,
            subtitle="%sGlobally" % group_prefix)
        pc_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_PARENT_PC, begin_ns, end_ns,
            subtitle="%sWith active children" % group_prefix)
        global_pc_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_PARENT_PC, begin_ns, end_ns,
            subtitle="%sGlobally" % group_prefix)
        ret = _StatsFreqTables()
        ret.per_parent_stats_table = duration_table
        ret.per_parent_count_table = count_table
        ret.global_duration_table = global_duration_table
        ret.global_count_table = global_count_table
        ret.per_parent_pc_table = pc_table
        ret.global_pc_table = global_pc_table

        for period in per_period_stats.keys():
            if self._analysis_conf._aggregate_by is not None and \
                    period not in self._analysis_conf._aggregate_by:
                continue
            ret.duration_values[period] = {}
            ret.count_values[period] = {}
            ret.pc_values[period] = {}
            ret.global_duration_values[period] = {}
            ret.global_count_values[period] = {}
            ret.global_pc_values[period] = {}
            for child in per_period_stats[period]._children.keys():
                if self._args.select is not None and \
                        child not in self._args.select:
                    continue
                c = per_period_stats[period]._children[child]
                if period not in c.parent_count.keys():
                    continue
                nogroup_c = not_grouped_per_period_stats[period]
                if per_period_stats[period].nr_periods == 0:
                    global_duration_avg = 0
                    global_count_avg = 0
                    duration_avg = 0
                    count_avg = 0
                    pc_avg = 0
                else:
                    global_duration_avg = c.total / \
                        nogroup_c.nr_periods
                    global_count_avg = c.total_count / \
                        nogroup_c.nr_periods
                    duration_avg = c.total / c.parent_count[period]
                    count_avg = c.total_count / c.parent_count[period]
                    pc_avg = c.total_pc / c.parent_count[period]
                    global_pc_avg = c.total_pc / \
                        nogroup_c.nr_periods

                if len(c.durations) > 2:
                    duration_stdev = mi.Duration(statistics.stdev(c.durations))
                    count_stdev = mi.Number(statistics.stdev(c.count_array))
                    pc_stdev = mi.Number(statistics.stdev(c.pc_array))
                else:
                    duration_stdev = mi.Unknown()
                    count_stdev = mi.Unknown()
                    pc_stdev = mi.Unknown()

                # Make temporary copies in case we need to reuse the
                # original table afterwards.
                global_durations = c.durations.copy()
                global_count_array = c.count_array.copy()
                global_pc_array = c.pc_array.copy()
                # Save the raw values if we need them for the frequency
                # distributions
                ret.duration_values[period][child] = c.durations.copy()
                ret.count_values[period][child] = c.count_array.copy()
                ret.pc_values[period][child] = c.pc_array.copy()
                ret.global_duration_values[period][child] = c.durations.copy()
                ret.global_count_values[period][child] = c.count_array.copy()
                ret.global_pc_values[period][child] = c.pc_array.copy()
                if c.parent_count[period] < \
                        nogroup_c.nr_periods:
                    global_min = 0
                    global_min_count = 0
                    global_min_pc = 0
                    for i in range(nogroup_c.nr_periods -
                                   c.parent_count[period]):
                        global_durations.append(0)
                        global_count_array.append(0)
                        global_pc_array.append(0)
                        ret.global_duration_values[period][child].append(0)
                        ret.global_count_values[period][child].append(0)
                        ret.global_pc_values[period][child].append(0)
                else:
                    global_min = c.min
                    global_min_count = c.min_count
                    global_min_pc = c.min_pc

                if nogroup_c.nr_periods > 2:
                    global_duration_stdev = mi.Duration(
                        statistics.stdev(global_durations))
                    global_count_stdev = mi.Number(
                        statistics.stdev(global_count_array))
                    global_pc_stdev = mi.Number(
                        statistics.stdev(global_pc_array))
                else:
                    global_duration_stdev = mi.Unknown()
                    global_count_stdev = mi.Unknown()
                    global_pc_stdev = mi.Unknown()

                duration_table.append_row(
                    name=mi.String(self._get_full_period_path(child)),
                    parent=mi.String(self._get_full_period_path(period)),
                    min_duration=mi.Duration(c.min),
                    avg_duration=mi.Duration(duration_avg),
                    max_duration=mi.Duration(c.max),
                    stdev_duration=duration_stdev,
                )

                count_table.append_row(
                    name=mi.String(self._get_full_period_path(child)),
                    parent=mi.String(self._get_full_period_path(period)),
                    min=mi.Number(c.min_count),
                    avg=mi.Number(count_avg),
                    max=mi.Number(c.max_count),
                    stdev=count_stdev,
                )

                pc_table.append_row(
                    name=mi.String(self._get_full_period_path(child)),
                    parent=mi.String(self._get_full_period_path(period)),
                    min=mi.Number(c.min_pc),
                    avg=mi.Number(pc_avg),
                    max=mi.Number(c.max_pc),
                    stdev=pc_stdev,
                )

                global_duration_table.append_row(
                    name=mi.String(self._get_full_period_path(child)),
                    parent=mi.String(self._get_full_period_path(period)),
                    min_duration=mi.Duration(global_min),
                    avg_duration=mi.Duration(global_duration_avg),
                    max_duration=mi.Duration(c.max),
                    stdev_duration=global_duration_stdev,
                )

                global_count_table.append_row(
                    name=mi.String(self._get_full_period_path(child)),
                    parent=mi.String(self._get_full_period_path(period)),
                    min=mi.Number(global_min_count),
                    avg=mi.Number(global_count_avg),
                    max=mi.Number(c.max_count),
                    stdev=global_count_stdev,
                )

                global_pc_table.append_row(
                    name=mi.String(self._get_full_period_path(child)),
                    parent=mi.String(self._get_full_period_path(period)),
                    min=mi.Number(global_min_pc),
                    avg=mi.Number(global_pc_avg),
                    max=mi.Number(c.max_pc),
                    stdev=global_pc_stdev,
                )
        return ret

    def _find_filtered_uniform_freq_values(self, per_period_group_stats):
        for period in per_period_group_stats.keys():
            table = per_period_group_stats[period]
            min, max, count, avg, total, total_list = \
                self._get_filtered_min_max_count_avg_total_values(
                    table.durations)
            min, max, step = self._find_uniform_freq_values(total_list, 1000,
                                                            'duration')
        # We only care about the last values
        return min, max, step

    def _get_grouped_by_period_stats_freq(self, begin_ns, end_ns,
                                          per_period_group_stats,
                                          group_prefix, freq_min,
                                          freq_max, freq_step):
        stats_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_PER_PERIOD_STATS,
                                         begin_ns, end_ns)
        freq_tables = []
        freq_tables_by_period_name = {}

        for period in per_period_group_stats.keys():
            if self._args.select is not None and \
                    period not in self._args.select:
                continue
            table = per_period_group_stats[period]
            min, max, count, avg, total, total_list = \
                self._get_filtered_min_max_count_avg_total_values(
                    table.durations)
            stdev = self._compute_period_duration_stdev_values(total_list)
            if math.isnan(stdev):
                stdev = mi.Unknown()
            else:
                stdev = mi.Duration(stdev)
            count = len(total_list)

            stats_table.append_row(
                name=mi.String(self._get_full_period_path(period)),
                count=mi.Number(count),
                min_duration=mi.Duration(min),
                avg_duration=mi.Duration(avg),
                max_duration=mi.Duration(max),
                stdev_duration=stdev,
                runtime=mi.Duration(total),
            )

            subtitle = '{}Duration of period: {}'.format(group_prefix, period)
            tmp_table = self._get_one_freq_result_table(
                self._MI_TABLE_CLASS_FREQ_DURATION, begin_ns, end_ns,
                freq_min, freq_max, freq_step, total_list, subtitle, 1000)
            freq_tables.append(tmp_table)
            freq_tables_by_period_name[period] = tmp_table
        return stats_table, freq_tables, freq_tables_by_period_name

    def _get_per_period_stats_result_table(self, begin_ns, end_ns,
                                           period_tree):
        stats_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_PER_PERIOD_STATS,
                                         begin_ns, end_ns)

        period_stats_list = []
        for parent in period_tree.keys():
            self._get_ordered_period_stats_list(parent, period_stats_list,
                                                period_tree[parent])

        for period_stats in period_stats_list:
            if not period_stats.period_list:
                continue

            if self._args.select is not None and \
                    period_stats.name not in self._args.select:
                continue

            if self._args.min_duration is None and \
                    self._args.max_duration is None:
                stdev = self._compute_period_duration_stdev(
                    period_stats.period_list)
                min = period_stats.min_duration
                max = period_stats.max_duration
                count = period_stats.count
                total = period_stats.total_duration
                if count > 0:
                    avg = period_stats.total_duration / \
                        period_stats.count
                else:
                    avg = 0
            else:
                min, max, count, avg, total, period_list = \
                    self._get_filtered_min_max_count_avg_total_flist(
                        period_stats.period_list)
                if count == 0:
                    continue
                stdev = self._compute_period_duration_stdev(period_list)

            if math.isnan(stdev):
                stdev = mi.Unknown()
            else:
                stdev = mi.Duration(stdev)

            stats_table.append_row(
                name=mi.String(self._get_full_period_path(period_stats.name)),
                count=mi.Number(count),
                min_duration=mi.Duration(min),
                avg_duration=mi.Duration(avg),
                max_duration=mi.Duration(max),
                stdev_duration=stdev,
                runtime=mi.Duration(total),
            )

        return stats_table

    def _get_one_aggregated_log_table(self, begin_ns, end_ns,
                                      per_parent_aggregated_dict, sub, top):
        table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_AGGREGATED_LOG, begin_ns, end_ns,
            subtitle=sub)
        for parent_period in per_parent_aggregated_dict.keys():
            for child_period in \
                    per_parent_aggregated_dict[parent_period].keys():
                child_period_list = \
                    per_parent_aggregated_dict[parent_period][child_period]
                min, max, count, avg, total, period_list = \
                    self._get_agg_filtered_min_max_count_avg_total_flist(
                        child_period_list)
                stdev = self._compute_period_agg_duration_stdev(
                    child_period_list)

                if math.isnan(stdev):
                    stdev = mi.Unknown()
                else:
                    stdev = mi.Duration(stdev)

                table.append_row(
                    parent_begin_ts=mi.Timestamp(
                        parent_period.start_ts),
                    parent_end_ts=mi.Timestamp(
                        parent_period.end_ts),
                    parent_name=mi.String(parent_period.name),
                    child_name=mi.String(self._get_full_period_path(
                        child_period)),
                    count=mi.Number(count),
                    min_duration=mi.Duration(min),
                    avg_duration=mi.Duration(avg),
                    max_duration=mi.Duration(max),
                    runtime=mi.Duration(total),
                    parent_captures=mi.String(parent_period.full_captures()),
                    stdev_duration=stdev,
                )
        return table

    def _get_aggregated_log_table(self, begin_ns, end_ns,
                                  per_parent_aggregated_dict,
                                  aggregated_groups,
                                  top=False):
        result_tables = []
        ag_list = ""
        for i in self._analysis_conf._select:
            if len(ag_list) == 0:
                ag_list = i
            else:
                ag_list = "%s, %s" % (ag_list, i)
        sub = "Aggregation of (%s) by %s" % (
            ag_list, self._analysis_conf._aggregate_by)

        if aggregated_groups is None:
            table = self._get_one_aggregated_log_table(
                begin_ns, end_ns, per_parent_aggregated_dict, sub, top)
            result_tables.append(table)
        else:
            for group in aggregated_groups.keys():
                group_sub = "%s, group: %s" % (sub, group)
                result_tables.append(self._get_one_aggregated_log_table(
                    begin_ns, end_ns, aggregated_groups[group], group_sub,
                    top))

        return result_tables

    def _fill_freq_result_table(self, period_list, stats, min_duration,
                                max_duration, step, freq_table):
        # The number of bins for the histogram
        resolution = self._args.freq_resolution

        if not self._args.freq_uniform:
            if self._args.min is not None:
                min_duration = self._args.min
            else:
                min_duration = stats.min

            if self._args.max is not None:
                max_duration = self._args.max
            else:
                max_duration = stats.max

            # ns to µs
            if min_duration is None:
                min_duration = 0
            else:
                min_duration /= 1000

            if max_duration is None:
                max_duration = 0
            else:
                max_duration /= 1000

            step = (max_duration - min_duration) / resolution

        if step == 0:
            return

        buckets = []
        counts = []

        for i in range(resolution):
            buckets.append(i * step)
            counts.append(0)

        for period_event in period_list:
            if not self._filter_event_duration(period_event):
                continue
            duration = period_event.duration / 1000
            index = int((duration - min_duration) / step)

            if index >= resolution:
                # special case for max value: put in last bucket (includes
                # its upper bound)
                if duration == max_duration:
                    counts[index - 1] += 1

                continue

            counts[index] += 1

        for index, count in enumerate(counts):
            lower_bound = index * step + min_duration
            upper_bound = (index + 1) * step + min_duration
            freq_table.append_row(
                lower=mi.Duration.from_us(lower_bound),
                upper=mi.Duration.from_us(upper_bound),
                count=mi.Number(count),
            )

    def _fill_freq_result_table_values(self, values, min_duration,
                                       max_duration, step, freq_table, ratio):
        # Differ from _fill_freq_result_table because we work directly with
        # a list of values instead of periods.

        # The number of bins for the histogram
        resolution = self._args.freq_resolution

        if not self._args.freq_uniform:
            if self._args.min is not None:
                min_duration = self._args.min
            else:
                min_duration = min(values) / ratio

            if self._args.max is not None:
                max_duration = self._args.max
            else:
                max_duration = max(values) / ratio

            # ns to µs
            if min_duration is None:
                min_duration = 0

            if max_duration is None:
                max_duration = 0

            step = (max_duration - min_duration) / resolution

        if step == 0:
            return

        buckets = []
        counts = []

        for i in range(resolution):
            buckets.append(i * step)
            counts.append(0)

        for v in values:
            if not self._filter_duration(v):
                continue
            duration = v / ratio
            index = int((duration - min_duration) / step)

            if index < 0:
                raise ValueError('Invalid range, duration=%s, min=%s, max=%s,'
                                 ' step=%s, resolution=%s' % (
                                     duration, min_duration, max_duration,
                                     step, resolution))

            if index >= resolution:
                # special case for max value: put in last bucket (includes
                # its upper bound)
                if duration == max_duration:
                    counts[index - 1] += 1

                continue

            counts[index] += 1

        for index, count in enumerate(counts):
            lower_bound = index * step + min_duration
            upper_bound = (index + 1) * step + min_duration
            freq_table.append_row(
                lower=mi.Duration.from_us(lower_bound),
                upper=mi.Duration.from_us(upper_bound),
                count=mi.Number(count),
            )

    def _get_total_freq_result_tables(self, begin_ns, end_ns):
        freq_tables = []
        period_lists, period_stats = self._get_total_period_lists_stats()
        min_duration = None
        max_duration = None
        step = None
        subtitle = 'All periods'

        if self._args.freq_uniform:
            durations = []

            for period_list in period_lists:
                for period_event in period_list:
                    if not self._filter_event_duration(period_event):
                        continue
                    durations.append(period_event.duration)

            min_duration, max_duration, step = \
                self._find_uniform_freq_values(durations)

        for period_list in period_lists:
            freq_table = \
                self._mi_create_result_table(
                    self._MI_TABLE_CLASS_FREQ_DURATION, begin_ns, end_ns,
                    subtitle)
            self._fill_freq_result_table(period_list, period_stats,
                                         min_duration, max_duration, step,
                                         freq_table)
            freq_tables.append(freq_table)

        return freq_tables

    def _get_per_group_freq_series_tables(self, begin_ns, end_ns,
                                          per_period_freq_group_by_tables,
                                          freq_tables_group_per_period_names):
        column_infos = [
            ('duration_lower', 'Duration (lower bound)', mi.Duration),
            ('duration_upper', 'Duration (upper bound)', mi.Duration),
        ]
        unique_group_names = {}
        per_period_tables = {}

        for group in freq_tables_group_per_period_names.keys():
            unique_group_names[group] = None
            column_infos.append((
                # urgh, need to sanitize for the namedtuple, only alnum
                '{}'.format(re.sub(r'\W+', '', group)),
                # subtitle:
                '{}'.format(group),
                mi.Number, 'count'))
            for period in freq_tables_group_per_period_names[group]:
                title = 'Period \'%s\' duration frequency distribution ' \
                        'per group' % (period)
                table_class = mi.TableClass(None, title, column_infos)
                result_table = mi.ResultTable(table_class, begin_ns, end_ns)
                per_period_tables[period] = result_table

        for period in per_period_tables.keys():
            for i in range(self._args.freq_resolution):
                first_group = next(iter(unique_group_names))
                row_tuple = [
                    freq_tables_group_per_period_names[first_group][period].
                    rows[i].lower,
                    freq_tables_group_per_period_names[first_group][period].
                    rows[i].upper]
                for group in freq_tables_group_per_period_names.keys():
                    group_table = freq_tables_group_per_period_names[group]
                    freq_row = group_table[period].rows[i]
                    row_tuple.append(freq_row.count)
                per_period_tables[period].append_row_tuple(tuple(row_tuple))
        return per_period_tables

    def _get_period_lists_stats(self):
        period_lists = {}
        period_stats = {}

        for period in self._analysis.all_period_stats.keys():
            period_list = self._analysis.all_period_stats[period].period_list

            if not period_list:
                continue
            if self._args.min_duration is None and \
                    self._args.max_duration is None:
                stdev = self._compute_period_duration_stdev(period_list)
                count = len(period_list)
                min = self._analysis.all_period_stats[period].min_duration
                max = self._analysis.all_period_stats[period].max_duration
                total = \
                    self._analysis.all_period_stats[period].total_duration
            else:
                min, max, count, avg, total, period_list = \
                    self._get_filtered_min_max_count_avg_total_flist(
                        period_list)
                stdev = self._compute_period_duration_stdev(period_list)

            period_stats[period] = _PeriodStats(
                count=count, min=min, max=max, stdev=stdev, total=total)
            period_lists[period] = period_list

        return period_lists, period_stats

    def _find_table_min_max_step(self, table, ratio, category):
        _min = None
        max = 0
        # Find the uniform freq values across all parent/child combinations
        for period in table.keys():
            for child in table[period].keys():
                tmp_min, tmp_max, tmp_step = \
                    self._find_uniform_freq_values(
                        table[period][child], ratio, category)
                if _min is None or tmp_min < _min:
                    _min = tmp_min
                if tmp_max > max:
                    max = tmp_max
        if _min is None:
            steps = 0
        else:
            steps = (max - _min) / self._args.freq_resolution
        return _min, max, steps

    def _find_uniform_values(self, tables):
        if not self._args.freq_uniform:
            return None, None, None, None, None, None, \
                None, None, None, None, None, None, \
                None, None, None, None, None, None

        duration_min, duration_max, duration_step = \
            self._find_table_min_max_step(tables.duration_values, 1000,
                                          'duration')
        global_duration_min, global_duration_max, global_duration_step = \
            self._find_table_min_max_step(tables.global_duration_values, 1000,
                                          'global_duration')

        count_min, count_max, count_step = \
            self._find_table_min_max_step(tables.count_values, 1, 'count')
        global_count_min, global_count_max, global_count_step = \
            self._find_table_min_max_step(tables.global_count_values, 1,
                                          'global_count')

        pc_min, pc_max, pc_step = \
            self._find_table_min_max_step(tables.pc_values, 1, 'pc')
        global_pc_min, global_pc_max, global_pc_step = \
            self._find_table_min_max_step(tables.global_pc_values, 1,
                                          'global_pc')

        return duration_min, duration_max, duration_step, \
            global_duration_min, global_duration_max, \
            global_duration_step, \
            count_min, count_max, count_step, \
            global_count_min, global_count_max, \
            global_count_step, \
            pc_min, pc_max, pc_step, \
            global_pc_min, global_pc_max, \
            global_pc_step

    def _get_one_freq_result_table(self, mi_class, begin_ns, end_ns,
                                   min, max, step, values,
                                   subtitle, ratio=1):
        freq_table = \
            self._mi_create_result_table(mi_class, begin_ns, end_ns, subtitle)
        self._fill_freq_result_table_values(values, min, max, step,
                                            freq_table, ratio)
        return freq_table

    def _get_per_parent_freq_result_table(self, begin_ns, end_ns,
                                          tables, group_prefix=''):
        duration_min, duration_max, duration_step, \
            global_duration_min, global_duration_max, \
            global_duration_step, \
            count_min, count_max, count_step, \
            global_count_min, global_count_max, \
            global_count_step, \
            pc_min, pc_max, pc_step, \
            global_pc_min, global_pc_max, \
            global_pc_step = self._find_uniform_values(tables)

        # sorted to get the same output order between runs
        for period in sorted(tables.duration_values.keys()):
            if self._analysis_conf._aggregate_by is not None and \
                    period not in self._analysis_conf._aggregate_by:
                continue
            for child in tables.duration_values[period].keys():
                if self._args.select is not None and \
                        child not in self._args.select:
                    continue
                subtitle = "%sDuration of %s per %s" % (
                    group_prefix,
                    self._get_full_period_path(child),
                    self._get_full_period_path(period))
                # ratio=1000 for ns -> us
                tables.per_parent_freq_tables.append(
                    self._get_one_freq_result_table(
                        self._MI_TABLE_CLASS_FREQ_DURATION,
                        begin_ns, end_ns, duration_min, duration_max,
                        duration_step, tables.duration_values[period][child],
                        subtitle, ratio=1000))

                subtitle = "%sNumber of %s per %s" % (
                    group_prefix,
                    self._get_full_period_path(child),
                    self._get_full_period_path(period))
                tables.per_parent_count_freq_tables.append(
                    self._get_one_freq_result_table(
                        self._MI_TABLE_CLASS_FREQ_COUNT,
                        begin_ns, end_ns, count_min, count_max, count_step,
                        tables.count_values[period][child],
                        subtitle))

                subtitle = "%sUsage ratio of %s per %s" % (
                    group_prefix,
                    self._get_full_period_path(child),
                    self._get_full_period_path(period))
                tables.per_parent_pc_freq_tables.append(
                    self._get_one_freq_result_table(
                        self._MI_TABLE_CLASS_FREQ_PC,
                        begin_ns, end_ns, pc_min, pc_max, pc_step,
                        tables.pc_values[period][child],
                        subtitle))

                subtitle = "%sGlobal duration of %s per %s" % (
                    group_prefix,
                    self._get_full_period_path(child),
                    self._get_full_period_path(period))
                # ratio=1000 for ns -> us
                tables.global_duration_freq_tables.append(
                    self._get_one_freq_result_table(
                        self._MI_TABLE_CLASS_FREQ_DURATION,
                        begin_ns, end_ns, global_duration_min,
                        global_duration_max, global_duration_step,
                        tables.global_duration_values[period][child],
                        subtitle, ratio=1000))

                subtitle = "%sGlobal number of %s per %s" % (
                    group_prefix,
                    self._get_full_period_path(child),
                    self._get_full_period_path(period))
                tables.global_count_freq_tables.append(
                    self._get_one_freq_result_table(
                        self._MI_TABLE_CLASS_FREQ_COUNT,
                        begin_ns, end_ns, global_count_min, global_count_max,
                        global_count_step,
                        tables.global_count_values[period][child],
                        subtitle))

                subtitle = "%sGlobal usage ratio of %s per %s" % (
                    group_prefix,
                    self._get_full_period_path(child),
                    self._get_full_period_path(period))
                tables.global_pc_freq_tables.append(
                    self._get_one_freq_result_table(
                        self._MI_TABLE_CLASS_FREQ_PC,
                        begin_ns, end_ns, global_pc_min, global_pc_max,
                        global_pc_step,
                        tables.global_pc_values[period][child],
                        subtitle))

    def _get_per_period_freq_result_tables(self, begin_ns, end_ns):
        freq_tables = []
        period_lists, period_stats = self._get_period_lists_stats()
        min_duration = None
        max_duration = None
        step = None

        if self._args.freq_uniform:
            durations = []

            for period_list in period_lists.values():
                for period_event in period_list:
                    if not self._filter_event_duration(period_event):
                        continue
                    durations.append(period_event.duration)

            min_duration, max_duration, step = \
                self._find_uniform_freq_values(durations)

        for period in sorted(period_stats.keys()):
            if self._args.select is not None and \
                    period not in self._args.select:
                continue
            period_list = period_lists[period]
            stats = period_stats[period]
            subtitle = 'Duration of period: {}'.format(period)
            freq_table = \
                self._mi_create_result_table(
                    self._MI_TABLE_CLASS_FREQ_DURATION, begin_ns, end_ns,
                    subtitle)
            self._fill_freq_result_table(period_list, stats,
                                         min_duration, max_duration, step,
                                         freq_table)
            freq_tables.append(freq_table)

        return freq_tables

    def _compute_period_duration_stdev_values(self, durations):
        period_durations = []
        for d in durations:
            if not self._filter_duration(d):
                continue
            period_durations.append(d)
        if len(period_durations) < 2:
            return float('nan')
        return statistics.stdev(period_durations)

    def _compute_period_duration_stdev(self, period_events):
        period_durations = []
        for period_event in period_events:
            if not self._filter_event_duration(period_event):
                continue
            period_durations.append(period_event.duration)
        if len(period_durations) < 2:
            return float('nan')
        return statistics.stdev(period_durations)

    def _compute_period_agg_duration_stdev(self, period_agg_events):
        period_durations = []
        for period_event in period_agg_events:
            if not self._filter_event_duration(period_event.event):
                continue
            period_durations.append(period_event.event.duration)
        if len(period_durations) < 2:
            return float('nan')
        return statistics.stdev(period_durations)

    def _pop_next_capture_string(self, begin_captures, end_captures):
        if len(begin_captures.keys()) > 0:
            b_key, b_value = begin_captures.popitem()
            b_string = '%s = %s' % (b_key, b_value)
        else:
            b_string = ''

        if len(end_captures.keys()) > 0:
            e_key, e_value = end_captures.popitem()
            e_string = '%s = %s' % (e_key, e_value)
        else:
            e_string = ''

        return b_string, e_string

    def _print_period_events(self, result_table):
        fmt = '[{:<18}, {:<18}] {:>15} {:<24} {:<35} {:<35}'
        fmt_captures = '{:<18} {:<18} {:>18} {:<24} {:<35} {:<35}'
        title_fmt = '{:<20} {:<19} {:>15} {:<24} {:<35} {:<35}'
        print()
        print(result_table.title)
        print(title_fmt.format('Begin', 'End', 'Duration (us)', 'Name',
                               'Begin capture', 'End capture'))
        for row in result_table.rows:
            begin_ts = row.begin_ts.value
            end_ts = row.end_ts.value
            duration = row.duration.value
            name = row.name.value
            if name is None:
                name = ''

            # Convert back the string to dict
            begin_captures = ast.literal_eval(row.begin_captures.value)
            # Order the dict based on keys to always get the same output
            if begin_captures is None:
                begin_captures = {}
            begin_captures = collections.OrderedDict(
                sorted(begin_captures.items()))
            end_captures = ast.literal_eval(row.end_captures.value)
            if end_captures is None:
                end_captures = {}
            end_captures = collections.OrderedDict(
                sorted(end_captures.items()))

            b_string, e_string = self._pop_next_capture_string(begin_captures,
                                                               end_captures)

            print(fmt.format(self._format_timestamp(begin_ts),
                             self._format_timestamp(end_ts),
                             '%0.03f' % (duration / 1000), name,
                             b_string, e_string))

            nr_lines = max(len(begin_captures.keys()),
                           len(end_captures.keys()))
            for i in range(nr_lines):
                b_string, e_string = self._pop_next_capture_string(
                    begin_captures, end_captures)
                print(fmt_captures.format('', '', '', '', b_string, e_string))

    def _print_aggregated_period_events(self, result_tables):
        fmt = '[{:<18}, {:<18}] {:>22} {:<15} [{:<18}, {:<18}] {:>22} ' \
            '{:<15} {:<35}'
#        fmt_captures = '{:<18} {:<18} {:>25} {:<15} {:<18} {:<25} {:>18} ' \
#            '{:<15} {:<35}'
        title_fmt = '{:<20} {:<19} {:>22} {:<15} {:<20} {:<19} {:>22} ' \
            '{:<15} {:<35}'
        for result_table in result_tables:
            print()
            print(result_table.title)
            print(result_table.subtitle)
            print(title_fmt.format('Parent begin', 'Parent end',
                                   'Parent duration (us)', 'Parent name',
                                   'Child begin', 'Child end',
                                   'Child duration (us)', 'Child name',
                                   'Captures'))
            for row in result_table.rows:
                parent_begin_ts = row.parent_begin_ts.value
                parent_end_ts = row.parent_end_ts.value
                parent_duration = row.parent_duration.value
                parent_name = row.parent_name.value
                child_begin_ts = row.child_begin_ts.value
                child_end_ts = row.child_end_ts.value
                child_duration = row.child_duration.value
                child_name = row.child_name.value

                # Convert back the string to list of tuple
                captures = ast.literal_eval(row.captures.value)
                # Order the dict based on keys to always get the same output
#                if captures is None:
#                    captures = []
#                    capture_str = ''
#                else:
#                    captures = sorted(captures, key=lambda x: x[0])
#                    captures.reverse()
#                    tmp = captures.pop()
#                    capture_str = "%s = %s" % (tmp[0], tmp[1])
                capture_str = ''
                for i in sorted(captures, key=lambda x: x[0]):
                    if len(capture_str) == 0:
                        capture_str = "%s = %s" % (i[0], i[1])
                    else:
                        capture_str = "%s, %s = %s" % (capture_str, i[0], i[1])

                print(fmt.format(self._format_timestamp(parent_begin_ts),
                                 self._format_timestamp(parent_end_ts),
                                 '%0.03f' % (parent_duration / 1000),
                                 parent_name,
                                 self._format_timestamp(child_begin_ts),
                                 self._format_timestamp(child_end_ts),
                                 '%0.03f' % (child_duration / 1000),
                                 child_name,
                                 capture_str))
#                for i in range(len(captures)):
#                    tmp = captures.pop()
#                    capture_str = "%s = %s" % (tmp[0], tmp[1])
#                    print(fmt_captures.format('', '', '', '', '', '', '', '',
#                                              capture_str))

    def _print_total_stats(self, stats_table):
        row_format = '{:<12} {:<12} {:<12} {:<12} {:<12}'
        header = row_format.format(
            'Count', 'Min', 'Avg', 'Max', 'Stdev'
        )

        if stats_table.rows:
            print()
            print(stats_table.title + ' (us)')
            print(header)
            for row in stats_table.rows:
                if type(row.stdev_duration) is mi.Unknown:
                    stdev_str = '?'
                else:
                    stdev_str = '%0.03f' % row.stdev_duration.to_us()

                row_str = row_format.format(
                    '%d' % row.count.value,
                    '%0.03f' % row.min_duration.to_us(),
                    '%0.03f' % row.avg_duration.to_us(),
                    '%0.03f' % row.max_duration.to_us(),
                    '%s' % stdev_str,
                )

                print(row_str)

    def _print_period_tree(self, period_tree, level):
        for parent in period_tree.keys():
            if level == 0:
                lines = ''
            else:
                lines = "%s|-- " % ((level - 1) * 4 * ' ')
            print("%s%s" % (lines, parent))
            for child in period_tree[parent]:
                self._print_period_tree(period_tree[parent], level + 1)

    def _print_per_period_stats(self, stats_table, period_tree):
        row_format = '{:<25} {:>8}  {:>12}  {:>12}  {:>12}  {:>12} {:>12}'
        header = row_format.format(
            'Period', 'Count', 'Min', 'Avg', 'Max', 'Stdev', 'Runtime'
        )

        print("Period tree:")
        self._print_period_tree(period_tree, 0)

        if stats_table.rows:
            print()
            print(stats_table.title + ' (us)')
            print(header)
            for row in stats_table.rows:
                if type(row.stdev_duration) is mi.Unknown:
                    stdev_str = '?'
                else:
                    stdev_str = '%0.03f' % row.stdev_duration.to_us()

                row_str = row_format.format(
                    '%s' % row.name,
                    '%d' % row.count.value,
                    '%0.03f' % row.min_duration.to_us(),
                    '%0.03f' % row.avg_duration.to_us(),
                    '%0.03f' % row.max_duration.to_us(),
                    '%s' % stdev_str,
                    '%0.03f' % row.runtime.to_us(),
                )

                print(row_str)

    def _print_per_parent_stats(self, table):
        row_format = '{:<25} {:<25}  {:>12}  {:>12}  {:>12}  {:>12}'
        header = row_format.format(
            'Period', 'Parent', 'Min', 'Avg', 'Max', 'Stdev'
        )
        if table.rows:
            print()
            print(table.title + ' (us)')
            print(table.subtitle)
            print(header)
            for row in table.rows:
                if type(row.stdev_duration) is mi.Unknown:
                    stdev_str = '?'
                else:
                    stdev_str = '%0.03f' % row.stdev_duration.to_us()
                row_str = row_format.format(
                    '%s' % row.name,
                    '%s' % row.parent,
                    '%0.03f' % row.min_duration.to_us(),
                    '%0.03f' % row.avg_duration.to_us(),
                    '%0.03f' % row.max_duration.to_us(),
                    '%s' % stdev_str,
                )
                print(row_str)

    def _print_per_parent_count(self, table):
        row_format = '{:<25} {:<25}  {:>12}  {:>12}  {:>12}  {:>12}'
        header = row_format.format(
            'Period', 'Parent', 'Min', 'Avg', 'Max', 'Stdev'
        )
        if table.rows:
            print()
            print(table.title)
            print(table.subtitle)
            print(header)
            for row in table.rows:
                if type(row.stdev) is mi.Unknown:
                    stdev_str = '?'
                else:
                    stdev_str = "%0.03f" % row.stdev.value
                row_str = row_format.format(
                    '%s' % row.name,
                    '%s' % row.parent,
                    '%d' % row.min.value,
                    '%0.03f' % row.avg.value,
                    '%d' % row.max.value,
                    '%s' % stdev_str,
                )
                print(row_str)

    def _print_per_parent_pc(self, table):
        row_format = '{:<25} {:<25}  {:>12}  {:>12}  {:>12}  {:>12}'
        header = row_format.format(
            'Period', 'Parent', 'Min', 'Avg', 'Max', 'Stdev'
        )
        if table.rows:
            print()
            print(table.title + ' (%)')
            print(table.subtitle)
            print(header)
            for row in table.rows:
                if type(row.stdev) is mi.Unknown:
                    stdev_str = '?'
                else:
                    stdev_str = "%0.03f" % row.stdev.value
                row_str = row_format.format(
                    '%s' % row.name,
                    '%s' % row.parent,
                    '%d' % row.min.value,
                    '%0.03f' % row.avg.value,
                    '%d' % row.max.value,
                    '%s' % stdev_str,
                )
                print(row_str)

    def _one_line_captures(self, capture_tuple_list):
        capture_str = None
        for item in capture_tuple_list:
            if capture_str is None:
                capture_str = "%s = %s" % (item[0], item[1])
                continue
            capture_str = "%s, %s = %s" % (capture_str, item[0], item[1])
        return capture_str

    def _print_aggregated_log(self, stats_tables):
        fmt = '[{:<18}, {:<18}] {:>18} {:<15} {:<24} {:>12} | {:>10} ' \
            '{:>12} {:>12} {:>12} {:>13} | {:>12}'
        title_fmt = '{:<20} {:<19} {:>18} {:<15} {:<24} {:>12} | {:>10} ' \
            '{:>12} {:>12} {:>13} {:>12} | {:>12}'
        high_title_fmt = '{:<35} Parent {:<32} | {:<35} | {:<25} ' \
            'Durations (us) {:<22} |'
        for stats_table in stats_tables:
            print()
            print(stats_table.title)
            print(stats_table.subtitle)
            print(high_title_fmt.format('', '', '', '', ''))
            print(title_fmt.format('Begin', 'End',
                                   'Duration (us)', 'Name',
                                   '| Child name', 'Count', 'Min', 'Avg',
                                   'Max', 'Stdev', 'Runtime',
                                   'Parent captures'))

            for row in stats_table.rows:
                parent_begin_ts = row.parent_begin_ts.value
                parent_end_ts = row.parent_end_ts.value
                parent_duration = parent_end_ts - parent_begin_ts
                parent_name = row.parent_name.value
                child_name = row.child_name.value
                captures = self._one_line_captures(row.parent_captures.value)
                if type(row.stdev_duration) is mi.Unknown:
                    stdev_str = '?'
                else:
                    stdev_str = '%0.03f' % row.stdev_duration.to_us()

                row_str = fmt.format(
                    self._format_timestamp(parent_begin_ts),
                    self._format_timestamp(parent_end_ts),
                    '%0.03f' % (parent_duration / 1000),
                    parent_name,
                    '| %s' % child_name,
                    '%d' % row.count.value,
                    '%0.03f' % row.min_duration.to_us(),
                    '%0.03f' % row.avg_duration.to_us(),
                    '%0.03f' % row.max_duration.to_us(),
                    '%s' % stdev_str,
                    '%0.03f' % row.runtime.to_us(),
                    '%s' % captures,
                )

                print(row_str)

    def _print_frequency_distribution(self, freq_table, unit):
        title_fmt = '{} - {}'

        graph = termgraph.FreqGraph(
            data=freq_table.rows,
            get_value=lambda row: row.count.value,
            get_lower_bound=lambda row: row.lower.to_us(),
            title=title_fmt.format(freq_table.title, freq_table.subtitle),
            unit=unit,
        )

        graph.print_graph()

    def _print_freq(self, freq_tables, unit):
        for freq_table in freq_tables:
            self._print_frequency_distribution(freq_table, unit)

    def _cleanup_period_name(self, name):
        # If a period name is given with its hierarchy, only keep the last
        # member, period names must be unique, so we don't need to scope them,
        # but since we output the periods with their full hierarchy, we have
        # to support it in entry as well.
        return name.split('/')[-1]

    def _validate_transform_args(self):
        args = self._args
        self._analysis_conf._group_by = {}
        self._analysis_conf._aggregate_by = None
        self._analysis_conf._select = []
        self._analysis_conf._order_by = None

        if args.group_by:
            for group in args.group_by.split(','):
                g = group.strip()
                if len(g) == 0:
                    continue
                _period_name = self._cleanup_period_name(g.split('.')[0])
                _period_field = g.split('.')[1]
                if _period_name not in \
                        self._analysis_conf._group_by.keys():
                    self._analysis_conf._group_by[_period_name] = []
                self._analysis_conf._group_by[_period_name]. \
                    append(_period_field)

        if args.order_by:
            if args.order_by not in ['time', 'hierarchy']:
                self._gen_error("Invalid order-by value")
            self._analysis_conf._order_by = args.order_by

        # TODO: check aggregation and group-by attributes are valid
        if args.select:
            for ag in args.select.split(','):
                self._analysis_conf._select.append(
                    self._cleanup_period_name(ag).strip())
        if args.aggregate_by is not None:
            self._analysis_conf._aggregate_by = self._cleanup_period_name(
                args.aggregate_by)

    def _add_arguments(self, ap):
        Command._add_min_max_args(ap)
        Command._add_freq_args(
            ap, help='Output statistics about periods durations')
        Command._add_top_args(ap, help='Output the top sched switch durations')
        Command._add_log_args(
            ap, help='Output the sched switches in chronological order')
        Command._add_stats_args(ap, help='Output sched switch statistics')
        ap.add_argument('--min-duration', type=float,
                        help='Filter out, periods shorter that duration '
                             '(usec)')
        ap.add_argument('--max-duration', type=float,
                        help='Filter out, periods longer than duration (usec)')
        ap.add_argument('--aggregate-by', type=str,
                        help='FIXME')
        ap.add_argument('--select', type=str,
                        help='FIXME')
        ap.add_argument('--group-by', type=str,
                        help='Present the results grouped by a list of fields'
                             '(period.captured_field'
                             '[, period.captured_field2])')
        ap.add_argument('--order-by', type=str,
                        help='hierarchy, time')


def _run(mi_mode):
    schedcmd = PeriodAnalysisCommand(mi_mode=mi_mode)
    schedcmd.run()


def _runstats(mi_mode):
    sys.argv.insert(1, '--stats')
    _run(mi_mode)


def _runlog(mi_mode):
    sys.argv.insert(1, '--log')
    _run(mi_mode)


def _runtop(mi_mode):
    sys.argv.insert(1, '--top')
    _run(mi_mode)


def _runfreq(mi_mode):
    sys.argv.insert(1, '--freq')
    _run(mi_mode)


def runstats():
    _runstats(mi_mode=False)


def runlog():
    _runlog(mi_mode=False)


def runtop():
    _runtop(mi_mode=False)


def runfreq():
    _runfreq(mi_mode=False)


def runstats_mi():
    _runstats(mi_mode=True)


def runlog_mi():
    _runlog(mi_mode=True)


def runtop_mi():
    _runtop(mi_mode=True)


def runfreq_mi():
    _runfreq(mi_mode=True)
